"""Gateway for MCP (Model Context Protocol) integration with Invariant."""

import sys
import subprocess
import json
import os
import select
import asyncio

from invariant_sdk.async_client import AsyncClient
from invariant_sdk.types.append_messages import AppendMessagesRequest
from invariant_sdk.types.push_traces import PushTracesRequest

from gateway.common.constants import (
    INVARIANT_GUARDRAILS_BLOCKED_MESSAGE,
    INVARIANT_GUARDRAILS_BLOCKED_TOOLS_MESSAGE,
    MCP_METHOD,
    MCP_CLIENT_INFO,
    MCP_PARAMS,
    MCP_SERVER_INFO,
    MCP_TOOL_CALL,
    MCP_LIST_TOOLS,
)
from gateway.common.guardrails import GuardrailAction
from gateway.common.request_context import RequestContext
from gateway.integrations.explorer import create_annotations_from_guardrails_errors
from gateway.integrations.guardrails import check_guardrails
from gateway.mcp.log import mcp_log, MCP_LOG_FILE, format_errors_in_response
from gateway.mcp.mcp_context import McpContext
from gateway.mcp.task_utils import run_task_in_background, run_task_sync
import getpass
import socket

UTF_8_ENCODING = "utf-8"
DEFAULT_API_URL = "https://explorer.invariantlabs.ai"


def user_and_host() -> str:
    username = getpass.getuser()
    hostname = socket.gethostname()

    return f"{username}@{hostname}"


def session_metadata(ctx: McpContext) -> dict:
    return {
        "session_id": ctx.local_session_id,
        "system_user": user_and_host(),
        "mcp_client": ctx.mcp_client_name,
        "mcp_server": ctx.mcp_server_name,
        "tools": ctx.tools,
        **(ctx.extra_metadata or {}),
    }


def write_as_utf8_bytes(data: dict) -> bytes:
    """Serializes dict to bytes using UTF-8 encoding."""
    return json.dumps(data).encode(UTF_8_ENCODING) + b"\n"


def deduplicate_annotations(ctx: McpContext, new_annotations: list) -> list:
    """Deduplicate new_annotations using the annotations in the context."""
    deduped_annotations = []
    for annotation in new_annotations:
        # Check if an annotation with the same content and address exists in ctx.annotations
        # TODO: Rely on the __eq__ method of the AnnotationCreate class directly via not in
        # to remove duplicates instead of using a custom logic.
        # This is a temporary solution until the Invariant SDK is updated.
        is_duplicate = False
        for ctx_annotation in ctx.annotations:
            if (
                annotation.content == ctx_annotation.content
                and annotation.address == ctx_annotation.address
                and annotation.extra_metadata == ctx_annotation.extra_metadata
            ):
                is_duplicate = True
                break

        if not is_duplicate:
            deduped_annotations.append(annotation)

    return deduped_annotations


def check_if_new_errors(ctx: McpContext, guardrails_result: dict) -> bool:
    """Checks if there are new errors in the guardrails result."""
    annotations = create_annotations_from_guardrails_errors(
        guardrails_result.get("errors", [])
    )
    for annotation in annotations:
        if annotation not in ctx.annotations:
            return True
    return False


async def get_guardrails_check_result(
    ctx: McpContext,
    message: dict,
    action: GuardrailAction = GuardrailAction.BLOCK,
) -> dict:
    """
    Check against guardrails of type action in an async manner.
    """
    # Skip if no guardrails are configured for this action
    if not (
        (ctx.guardrails.blocking_guardrails and action == GuardrailAction.BLOCK)
        or (ctx.guardrails.logging_guardrails and action == GuardrailAction.LOG)
    ):
        return {}

    # Prepare context and select appropriate guardrails
    context = RequestContext.create(
        request_json={},
        dataset_name=ctx.explorer_dataset,
        invariant_authorization="Bearer " + os.getenv("INVARIANT_API_KEY"),
        guardrails=ctx.guardrails,
    )

    guardrails_to_check = (
        ctx.guardrails.blocking_guardrails
        if action == GuardrailAction.BLOCK
        else ctx.guardrails.logging_guardrails
    )

    # Run check_guardrails asynchronously
    return await check_guardrails(
        messages=ctx.trace + [message],
        guardrails=guardrails_to_check,
        context=context,
    )


async def append_and_push_trace(
    ctx: McpContext, message: dict, guardrails_result: dict
) -> None:
    """
    Append a message to the trace if it exists or create a new one
    and push it to the Invariant Explorer.
    """
    annotations = []
    if guardrails_result and guardrails_result.get("errors", []):
        annotations = create_annotations_from_guardrails_errors(
            guardrails_result["errors"]
        )

    if ctx.guardrails.logging_guardrails:
        logging_guardrails_check_result = await get_guardrails_check_result(
            ctx, message, action=GuardrailAction.LOG
        )
        if logging_guardrails_check_result and logging_guardrails_check_result.get(
            "errors", []
        ):
            annotations.extend(
                create_annotations_from_guardrails_errors(
                    logging_guardrails_check_result["errors"]
                )
            )
    deduplicated_annotations = deduplicate_annotations(ctx, annotations)

    try:
        # If the trace_id is None, create a new trace with the messages.
        # Otherwise, append the message to the existing trace.
        client = AsyncClient(
            api_url=os.getenv("INVARIANT_API_URL", DEFAULT_API_URL),
        )

        if ctx.trace_id is None:
            ctx.trace.append(message)

            # default metadata
            metadata = {"source": "mcp"}
            # include MCP session metadata
            metadata.update(session_metadata(ctx))

            response = await client.push_trace(
                PushTracesRequest(
                    messages=[ctx.trace],
                    dataset=ctx.explorer_dataset,
                    metadata=[metadata],
                    annotations=[deduplicated_annotations],
                )
            )
            ctx.trace_id = response.id[0]
            ctx.last_trace_length = len(ctx.trace)
            ctx.annotations.extend(deduplicated_annotations)
        else:
            ctx.trace.append(message)
            response = await client.append_messages(
                AppendMessagesRequest(
                    trace_id=ctx.trace_id,
                    messages=ctx.trace[ctx.last_trace_length :],
                    annotations=deduplicated_annotations,
                )
            )
            ctx.last_trace_length = len(ctx.trace)
            ctx.annotations.extend(deduplicated_annotations)
    except Exception as e:
        mcp_log("[ERROR] Error pushing trace in append_and_push_trace:", e)


async def get_guardrails_check_result(
    ctx: McpContext,
    message: dict,
    action: GuardrailAction = GuardrailAction.BLOCK,
) -> dict:
    """
    Check against guardrails of type action.
    Works in both sync and async contexts by always using a dedicated thread.
    """
    # Skip if no guardrails are configured for this action
    if not (
        (ctx.guardrails.blocking_guardrails and action == GuardrailAction.BLOCK)
        or (ctx.guardrails.logging_guardrails and action == GuardrailAction.LOG)
    ):
        return {}

    # Prepare context and select appropriate guardrails
    context = RequestContext.create(
        request_json={},
        dataset_name=ctx.explorer_dataset,
        invariant_authorization="Bearer " + os.getenv("INVARIANT_API_KEY"),
        guardrails=ctx.guardrails,
        guardrails_parameters={"metadata": session_metadata(ctx), "action": action},
    )

    guardrails_to_check = (
        ctx.guardrails.blocking_guardrails
        if action == GuardrailAction.BLOCK
        else ctx.guardrails.logging_guardrails
    )

    return run_task_sync(
        check_guardrails,
        messages=ctx.trace + [message],
        guardrails=guardrails_to_check,
        context=context,
    )


def json_rpc_error_response(
    id_value: str | int, error_message: str, response_type: str = "error"
) -> dict:
    """
    Create a JSON-RPC error response with either error object or content format.

    Args:
        id_value: The ID of the JSON-RPC request
        error_message: The error message to include
        response_type: Either "error" or "content" to determine response format

    Returns:
        A properly formatted JSON-RPC response dictionary
    """
    base_response = {
        "jsonrpc": "2.0",
        "id": id_value,
    }

    if response_type == "error":
        base_response["error"] = {
            "code": -32600,
            "message": error_message,
        }
    else:
        base_response["result"] = {
            "content": [
                {
                    "type": "text",
                    "text": error_message,
                }
            ]
        }

    return base_response


async def hook_tool_call(ctx: McpContext, request: dict) -> tuple[dict, bool]:
    """
    Hook function to intercept tool calls.

    If the request is blocked, it returns a tuple with a message explaining the block
    and a flag indicating the request was blocked.
    Otherwise it returns the original request and a flag indicating it was not blocked.
    """
    tool_call = {
        "id": f"call_{request.get('id')}",
        "type": "function",
        "function": {
            "name": request["params"]["name"],
            "arguments": request["params"]["arguments"],
        },
    }

    message = {"role": "assistant", "content": "", "tool_calls": [tool_call]}

    # Check for blocking guardrails
    guardrailing_result = await get_guardrails_check_result(
        ctx, message, action=GuardrailAction.BLOCK
    )

    # If the request is blocked, return a message indicating the block reason.
    if (
        guardrailing_result
        and guardrailing_result.get("errors", [])
        and check_if_new_errors(ctx, guardrailing_result)
    ):
        if ctx.push_explorer:
            await append_and_push_trace(ctx, message, guardrailing_result)
        else:
            ctx.trace.append(message)

        return json_rpc_error_response(
            request.get("id"),
            INVARIANT_GUARDRAILS_BLOCKED_MESSAGE % guardrailing_result["errors"],
            response_type=ctx.failure_response_format,
        ), True

    # Add the message to the trace
    ctx.trace.append(message)
    return request, False


async def hook_tool_result(ctx: McpContext, result: dict) -> dict:
    """
    Hook function to intercept tool results.
    Returns the potentially modified result.
    """
    method = ctx.id_to_method_mapping.get(result.get("id"))
    call_id = f"call_{result.get('id')}"

    # Safely handle result object
    result_obj = result.get("result", {})
    if isinstance(result_obj, dict) and MCP_SERVER_INFO in result_obj:
        ctx.mcp_server_name = result_obj.get(MCP_SERVER_INFO, {}).get("name", "")

    if method is None:
        return result
    elif method == MCP_TOOL_CALL:
        message = {
            "role": "tool",
            "content": result_obj.get("content"),
            "error": result_obj.get("error"),
            "tool_call_id": call_id
        }
        # Check for blocking guardrails
        guardrailing_result = await get_guardrails_check_result(
            ctx, message, action=GuardrailAction.BLOCK
        )

        if guardrailing_result and guardrailing_result.get("errors", []):
            result = json_rpc_error_response(
                result.get("id"),
                INVARIANT_GUARDRAILS_BLOCKED_MESSAGE
                % format_errors_in_response(guardrailing_result["errors"]),
                response_type=ctx.failure_response_format,  # Using content type as that's what the original code used
            )

        if ctx.push_explorer:
            await append_and_push_trace(ctx, message, guardrailing_result)
        else:
            ctx.trace.append(message)

        return result
    elif method == MCP_LIST_TOOLS:
        ctx.tools = result_obj.get("tools")
        message = {
            "role": "tool",
            "content": json.dumps(result.get("result").get("tools")),
            "tool_call_id": call_id,
        }
        # next validate it with guardrails
        guardrailing_result = await get_guardrails_check_result(
            ctx, message, action=GuardrailAction.BLOCK
        )
        if guardrailing_result and guardrailing_result.get("errors", []):
            result["result"]["tools"] = [
                {
                    "name": "blocked_" + tool["name"],
                    "description": INVARIANT_GUARDRAILS_BLOCKED_TOOLS_MESSAGE
                    % format_errors_in_response(guardrailing_result["errors"]),
                    # no parameters
                    "inputSchema": {
                        "properties": {},
                        "required": [],
                        "title": "invariant_mcp_server_blockedArguments",
                        "type": "object",
                    },
                    "annotations": {
                        "title": "This tool was blocked by security guardrails.",
                    },
                }
                for tool in result["result"]["tools"]
            ]

        # add it to the session trace (and run logging guardrails)
        if ctx.push_explorer:
            await append_and_push_trace(ctx, message, guardrailing_result)
        else:
            ctx.trace.append(message)

        return result
    else:
        return result


async def stream_and_forward_stdout(
    mcp_process: subprocess.Popen, ctx: McpContext
) -> None:
    """Read from the mcp_process stdout, apply guardrails and forward to sys.stdout"""
    loop = asyncio.get_event_loop()

    while True:
        line = await loop.run_in_executor(None, mcp_process.stdout.readline)
        if not line:
            break

        try:
            # Process complete JSON lines
            line_str = line.decode(UTF_8_ENCODING).strip()
            if not line_str:
                continue

            if ctx.verbose:
                mcp_log(f"[INFO] server -> client: {line_str}")

            parsed_json = json.loads(line_str)
            processed_json = await hook_tool_result(ctx, parsed_json)

            # Write and flush immediately
            sys.stdout.buffer.write(write_as_utf8_bytes(processed_json))
            sys.stdout.buffer.flush()

        except Exception as e:
            import traceback
            mcp_log(traceback.format_exc())
            mcp_log(f"[ERROR] Error in stream_and_forward_stdout: {str(e)}")
            if line:
                mcp_log(f"[ERROR] Problematic line causing error: {line[:200]}...")


async def stream_and_forward_stderr(
    mcp_process: subprocess.Popen, read_chunk_size: int = 10
) -> None:
    """Read from the mcp_process stderr and write to sys.stderr"""
    loop = asyncio.get_event_loop()

    while True:
        # Read chunks asynchronously
        chunk = await loop.run_in_executor(
            None, lambda: mcp_process.stderr.read(read_chunk_size)
        )

        MCP_LOG_FILE.buffer.write(chunk)
        MCP_LOG_FILE.buffer.flush()


async def process_line(
    ctx: McpContext, mcp_process: subprocess.Popen, line: bytes
) -> None:
    if ctx.verbose:
        mcp_log(f"[INFO] client -> server: {line}")

    # Try to decode and parse as JSON to check for tool calls
    try:
        text = line.decode(UTF_8_ENCODING)
        parsed_json = json.loads(text)
    except json.JSONDecodeError as je:
        mcp_log(f"[ERROR] JSON decode error in run_stdio_input_loop: {str(je)}")
        mcp_log(f"[ERROR] Problematic line: {line[:200]}...")
        return

    if parsed_json.get(MCP_METHOD) is not None:
        ctx.id_to_method_mapping[parsed_json.get("id")] = parsed_json.get(
            MCP_METHOD
        )
    if "params" in parsed_json and "clientInfo" in parsed_json.get("params"):
        ctx.mcp_client_name = (
            parsed_json.get("params").get("clientInfo").get("name", "")
        )

    # Check if this is a tool call request
    if parsed_json.get(MCP_METHOD) == MCP_TOOL_CALL:
        # Refresh guardrails
        run_task_sync(ctx.load_guardrails)

        # Intercept and potentially block modify the request
        hook_tool_call_result, is_blocked = await hook_tool_call(ctx, parsed_json)
        if not is_blocked:
            # If blocked, hook_tool_call_result contains the original request.
            # Forward the request to the MCP process.
            # It will handle the request and return a response.
            mcp_process.stdin.write(write_as_utf8_bytes(hook_tool_call_result))
            mcp_process.stdin.flush()
        else:
            # If blocked, hook_tool_call_result contains the block message.
            # Forward the block message result back to the caller.
            # The original request is not passed to the MCP process.
            sys.stdout.buffer.write(write_as_utf8_bytes(hook_tool_call_result))
            sys.stdout.buffer.flush()
        return
    else:
        # pass through the request to the MCP process

        # for list_tools, extend the trace by a tool call
        if parsed_json.get(MCP_METHOD) == MCP_LIST_TOOLS:
            # Refresh guardrails
            run_task_sync(ctx.load_guardrails)

            # mcp_message_{}
            ctx.trace.append(
                {
                    "role": "assistant",
                    "content": "",
                    "tool_calls": [
                        {
                            "id": f"call_{parsed_json.get('id')}",
                            "type": "function",
                            "function": {
                                "name": "tools/list",
                                "arguments": {},
                            },
                        }
                    ],
                }
            )
        mcp_process.stdin.write(write_as_utf8_bytes(parsed_json))
        mcp_process.stdin.flush()


async def run_stdio_input_loop(
    ctx: McpContext,
    mcp_process: subprocess.Popen,
    stdout_task: asyncio.Task,
    stderr_task: asyncio.Task,
) -> None:
    """Handle standard input, intercept call and forward requests to mcp_process stdin."""
    loop = asyncio.get_event_loop()
    stdin_fd = sys.stdin.fileno()
    buffer = b""

    # Set stdin to non-blocking mode
    os.set_blocking(stdin_fd, False)

    try:
        while True:
            # Check for input using select
            ready, _, _ = await loop.run_in_executor(
                None, lambda: select.select([stdin_fd], [], [], 0.1)
            )

            if not ready:
                # No input available, yield to other tasks
                await asyncio.sleep(0.01)
                continue

            # Read available data
            chunk = await loop.run_in_executor(None, lambda: os.read(stdin_fd, 4096))
            if not chunk:
                break  # EOF

            buffer += chunk

            # Process complete lines
            while b"\n" in buffer:
                line, buffer = buffer.split(b"\n", 1)
                if not line:
                    continue

                await process_line(ctx, mcp_process, line)
    except (BrokenPipeError, KeyboardInterrupt):
        # Broken pipe = client disappeared, just start shutdown
        mcp_log("Client disconnected or keyboard interrupt")
    finally:
        # Close stdin
        if mcp_process.stdin:
            mcp_process.stdin.close()

        # Process any remaining data
        while b"\n" in buffer:
            line, buffer = buffer.split(b"\n", 1)
            if line:
                await process_line(ctx, mcp_process, line)

        # Terminate process if needed
        if mcp_process.poll() is None:
            mcp_process.terminate()
            try:
                await asyncio.wait_for(
                    loop.run_in_executor(None, mcp_process.wait), timeout=2
                )
            except asyncio.TimeoutError:
                mcp_process.kill()

        # Cancel I/O tasks
        stdout_task.cancel()
        stderr_task.cancel()

        # Final flush
        sys.stdout.flush()


def split_args(args: list[str] = None) -> tuple[list[str], list[str]]:
    """
    Splits CLI arguments into two parts:
    1. Arguments intended for the MCP gateway (everything before `--exec`)
    2. Arguments for the underlying MCP server (everything after `--exec`)

    Parameters:
        args (list[str]): The list of CLI arguments.

    Returns:
        Tuple[list[str], list[str]]: A tuple containing (mcp_gateway_args, mcp_server_command_args)
    """
    if not args:
        mcp_log("[ERROR] No arguments provided.")
        sys.exit(1)

    try:
        exec_index = args.index("--exec")
    except ValueError:
        mcp_log("[ERROR] '--exec' flag not found in arguments.")
        sys.exit(1)

    mcp_gateway_args = args[:exec_index]
    mcp_server_command_args = args[exec_index + 1 :]

    if not mcp_server_command_args:
        mcp_log("[ERROR] No arguments provided after '--exec'.")
        sys.exit(1)

    return mcp_gateway_args, mcp_server_command_args


async def execute(args: list[str] = None):
    """Main function to execute the MCP gateway."""
    if "INVARIANT_API_KEY" not in os.environ:
        mcp_log("[ERROR] INVARIANT_API_KEY environment variable is not set.")
        sys.exit(1)

    mcp_log("[INFO] Running with Python version:", sys.version)

    mcp_gateway_args, mcp_server_command_args = split_args(args)
    ctx = McpContext(mcp_gateway_args)

    mcp_process = subprocess.Popen(
        mcp_server_command_args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=0,
    )

    # Start async tasks for stdout and stderr
    stdout_task = asyncio.create_task(stream_and_forward_stdout(mcp_process, ctx))
    stderr_task = asyncio.create_task(stream_and_forward_stderr(mcp_process))

    # Handle forwarding stdin and intercept tool calls
    await run_stdio_input_loop(ctx, mcp_process, stdout_task, stderr_task)



"""Converts the request and response formats from Gemini to Invariant API format."""

def convert_request(request: dict) -> list[dict]:
    """Converts the request from Gemini API to Invariant API format."""
    openai_messages = []

    if "systemInstruction" in request:
        system_content = " ".join(
            part.get("text", "")
            for part in request["systemInstruction"].get("parts", [])
        )
        openai_messages.append({"role": "system", "content": system_content})

    if "contents" in request:
        for content in request["contents"]:
            role = content.get("role", "")

            if role == "user":
                message_content = []
                for part in content.get("parts", []):
                    if "text" in part:
                        message_content.append({"type": "text", "text": part["text"]})
                    elif "inlineData" in part:
                        # TODO: Handle other types of inline data.
                        # Currently, only images are supported.
                        # Gemini’s API returns URL-safe base64 (uses _ and -).
                        # We need to convert it to standard base64 (uses + and /).
                        inline_data = (
                            part["inlineData"]["data"]
                            .replace("-", "+")
                            .replace("_", "/")
                        )
                        message_content.append(
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:{part['inlineData']['mime_type']};base64,{inline_data}"
                                },
                            }
                        )
                    elif "functionResponse" in part:
                        result = part["functionResponse"]["response"].get("result", {})
                        # TODO: Fix this once Explorer rendering is fixed.
                        if not isinstance(result, dict):
                            result = str(result)
                        openai_messages.append(
                            {
                                "role": "tool",
                                "tool_name": part["functionResponse"]["name"],
                                "content": result,
                            }
                        )
                if message_content:
                    openai_messages.append({"role": "user", "content": message_content})

            elif role == "model":
                for part in content.get("parts", []):
                    if "text" in part:
                        openai_messages.append(
                            {"role": "assistant", "content": part["text"]}
                        )
                    elif "functionCall" in part:
                        openai_messages.append(
                            {
                                "role": "assistant",
                                "tool_calls": [
                                    {
                                        "type": "function",
                                        "function": {
                                            "name": part["functionCall"]["name"],
                                            "arguments": part["functionCall"].get(
                                                "args", {}
                                            ),
                                        },
                                    }
                                ],
                            }
                        )

    return openai_messages


def convert_response(response: dict) -> list[dict]:
    """Converts the response from Gemini API to Invariant API format."""
    openai_messages = []

    if "candidates" in response:
        for candidate in response["candidates"]:
            candidate_content = candidate.get("content", {})
            role = candidate_content.get("role", "")
            if role == "model":
                for part in candidate_content.get("parts", []):
                    if "text" in part:
                        openai_messages.append(
                            {"role": "assistant", "content": part["text"]}
                        )
                    elif "functionCall" in part:
                        openai_messages.append(
                            {
                                "role": "assistant",
                                "tool_calls": [
                                    {
                                        "type": "function",
                                        "function": {
                                            "name": part["functionCall"]["name"],
                                            "arguments": part["functionCall"].get(
                                                "args", {}
                                            ),
                                        },
                                    }
                                ],
                            }
                        )

    return openai_messages



"""Utility functions for the Invariant explorer."""

import os
import json

from typing import Any, Dict, List
from fastapi import HTTPException

from gateway.common.guardrails import GuardrailRuleSet, Guardrail, GuardrailAction
from invariant_sdk.async_client import AsyncClient
from invariant_sdk.types.push_traces import PushTracesRequest, PushTracesResponse
from invariant_sdk.types.annotations import AnnotationCreate

import httpx

DEFAULT_API_URL = "https://explorer.invariantlabs.ai"


def create_annotations_from_guardrails_errors(
    guardrails_errors: List[dict],
) -> List[AnnotationCreate]:
    """Create Explorer annotations from the guardrails errors."""
    annotations = []

    def _pick_most_specific_ranges(ranges: list[str]) -> list[str]:
        """
        Remove redundant prefixes from the list of ranges.

        If the ranges are ['messages.2', 'messages.2.content:25-30', 'messages.2.content']
        then this returns ['messages.2.content:25-30'].

        This picks the most specific subset of the ranges and removes the rest. If some
        range is a proper prefix of another range, it is removed.
        """
        ranges = sorted(ranges, key=len)
        result = []

        for i, s in enumerate(ranges):
            is_prefix = False
            for t in ranges[i + 1 :]:
                if t.startswith(s) and t != s:
                    is_prefix = True
                    break
            if not is_prefix:
                result.append(s)

        return result

    for error in guardrails_errors:
        content = error.get("args")[0]
        filtered_ranges = _pick_most_specific_ranges(list(error.get("ranges", [])))
        for r in filtered_ranges:
            annotations.append(
                AnnotationCreate(
                    content=content,
                    address=r,
                    extra_metadata={
                        "source": "guardrails-error",
                        # if included in error, also include information about guardrail source
                        **(
                            {"guardrail": error.get("guardrail")}
                            if error.get("guardrail")
                            else {}
                        ),
                    },
                )
            )
    # Remove duplicates
    # TODO: Rely on the __eq__ and __hash__ methods of the AnnotationCreate class
    # to remove duplicates instead of using a custom function.
    # This is a temporary solution until the Invariant SDK is updated.
    return remove_duplicates(annotations)


def remove_duplicates(annotations: List[AnnotationCreate]) -> List[AnnotationCreate]:
    """
    Remove duplicate annotations based on content, address, and extra_metadata.

    Two annotations are considered duplicates if they have the same content,
    address, and extra_metadata.
    """
    unique_annotations = []
    seen = set()

    for annotation in annotations:
        # Convert the entire extra_metadata dict to a JSON string
        # This creates a hashable representation regardless of nested content
        metadata_str = json.dumps(annotation.extra_metadata, sort_keys=True)

        # Create a unique identifier using all three fields
        unique_key = (annotation.content, annotation.address, metadata_str)

        if unique_key not in seen:
            seen.add(unique_key)
            unique_annotations.append(annotation)

    return unique_annotations


def get_explorer_api_url() -> str:
    """Get the Invariant Explorer API URL from the environment variable."""
    return os.getenv("INVARIANT_API_URL", DEFAULT_API_URL)


async def push_trace(
    messages: List[List[Dict[str, Any]]],
    dataset_name: str,
    invariant_authorization: str,
    annotations: List[List[AnnotationCreate]] = None,
    metadata: List[Dict[str, Any]] = None,
) -> PushTracesResponse:
    """Pushes traces to the dataset on the Invariant Explorer.

    If a dataset with the given name does not exist, it will be created.

    Args:
        messages (List[List[Dict[str, Any]]]): List of messages to push.
        dataset_name (str): Name of the dataset.
        invariant_authorization (str): Value of the
                                       invariant-authorization header.

    Returns:
        PushTracesResponse: Response containing the trace ID details.
    """
    # Remove any None values from the messages
    update_messages = [
        [{k: v for k, v in msg.items() if v is not None} for msg in msg_list]
        for msg_list in messages
    ]
    request = PushTracesRequest(
        messages=update_messages,
        annotations=annotations,
        dataset=dataset_name,
        metadata=metadata,
    )
    client = AsyncClient(
        api_url=get_explorer_api_url().rstrip("/"),
        api_key=invariant_authorization.split("Bearer ")[1],
    )
    try:
        return await client.push_trace(request)
    except Exception as e:
        print(f"Failed to push trace: {e}")
        return {"error": str(e)}


async def fetch_guardrails_from_explorer(
    dataset_name: str,
    invariant_authorization: str,
    client_name: str | None = None,
    server_name: str | None = None,
) -> GuardrailRuleSet:
    """Get the guardrails for the dataset.

    Returns:
        GuardrailRuleSet: The guardrails for the dataset grouped by their action.
    """

    # TODO: Implement a single API in explorer backend which can return
    # dataset details without requiring a username.

    client = httpx.AsyncClient(
        base_url=get_explorer_api_url().rstrip("/"),
        headers={
            "Authorization": invariant_authorization,
        },
    )

    # Get the user details.
    user_info_response = await client.get("/api/v1/user/identity", timeout=5)
    if user_info_response.status_code == 401:
        raise HTTPException(
            status_code=401,
            detail="Invalid Invariant API key. Please check your API key.",
        )
    elif user_info_response.status_code != 200:
        raise ValueError(
            f"Failed to get user details from Explorer: {user_info_response.status_code}, {user_info_response.text}"
        )
    user_details = user_info_response.json()
    username = user_details["username"]

    # Get the dataset policies.
    policies_response = await client.get(
        f"/api/v1/dataset/byuser/{username}/{dataset_name}/policy",
        params={
            **({"client_name": client_name} if client_name else {}),
            **({"server_name": server_name} if server_name else {}),
        },
    )
    if policies_response.status_code != 200:
        if policies_response.status_code == 404:
            # If the dataset does not exist, return empty guardrails.
            return GuardrailRuleSet(
                blocking_guardrails=[],
                logging_guardrails=[],
            )
        raise ValueError(
            f"Failed to get dataset details from Explorer: {policies_response.status_code}, {policies_response.text}"
        )
    policies_details = policies_response.json()
    guardrails = policies_details.get("policies", [])

    blocking_guardrails = []
    logging_guardrails = []
    for g in guardrails:
        action = g["action"]

        if not g["enabled"]:
            # Skip guardrails that are not enabled.
            continue

        if action not in (GuardrailAction.BLOCK, GuardrailAction.LOG):
            print("[Warning] Skipping unknown guardrail action: ", action)
            continue

        guardrail = Guardrail(
            id=g["id"],
            name=g["name"],
            content=g["content"],
            action=GuardrailAction(action),
        )

        if action == GuardrailAction.BLOCK:
            blocking_guardrails.append(guardrail)
        else:
            logging_guardrails.append(guardrail)

    return GuardrailRuleSet(
        blocking_guardrails=blocking_guardrails,
        logging_guardrails=logging_guardrails,
    )



"""Gateway service to forward requests to the MCP SSE servers"""

import asyncio
import json
import re
import os
from typing import Tuple

import httpx
from httpx_sse import aconnect_sse, ServerSentEvent
from fastapi import APIRouter, HTTPException, Request, Response
from fastapi.responses import StreamingResponse

from gateway.common.constants import (
    CLIENT_TIMEOUT,
    INVARIANT_GUARDRAILS_BLOCKED_MESSAGE,
    INVARIANT_GUARDRAILS_BLOCKED_TOOLS_MESSAGE,
    MCP_METHOD,
    MCP_TOOL_CALL,
    MCP_LIST_TOOLS,
    MCP_PARAMS,
    MCP_RESULT,
    MCP_SERVER_INFO,
    MCP_CLIENT_INFO,
)
from gateway.common.guardrails import GuardrailAction
from gateway.common.mcp_sessions_manager import (
    McpSessionsManager,
    SseHeaderAttributes,
)
from gateway.mcp.log import format_errors_in_response
from gateway.integrations.explorer import create_annotations_from_guardrails_errors

MCP_SERVER_POST_HEADERS = {
    "connection",
    "accept",
    "content-length",
    "content-type",
}
MCP_SERVER_SSE_HEADERS = {
    "connection",
    "accept",
    "cache-control",
}

gateway = APIRouter()
session_store = McpSessionsManager()


@gateway.post("/mcp/messages/")
@gateway.post("/mcp/sse/messages/")
async def mcp_post_gateway(
    request: Request,
) -> Response:
    """Proxy calls to the MCP Server tools"""
    query_params = dict(request.query_params)
    if not query_params.get("session_id"):
        return HTTPException(
            status_code=400,
            detail="Missing 'session_id' query parameter",
        )
    if not session_store.session_exists(query_params.get("session_id")):
        return HTTPException(
            status_code=400,
            detail="Session does not exist",
        )
    if not request.headers.get("mcp-server-base-url"):
        return HTTPException(
            status_code=400,
            detail="Missing 'mcp-server-base-url' header",
        )

    session_id = query_params.get("session_id")
    mcp_server_messages_endpoint = (
        _convert_localhost_to_docker_host(request.headers.get("mcp-server-base-url"))
        + "/messages/?"
        + session_id
    )
    request_body_bytes = await request.body()
    request_json = json.loads(request_body_bytes)
    session = session_store.get_session(session_id)
    if request_json.get(MCP_METHOD) and request_json.get("id"):
        session.id_to_method_mapping[request_json.get("id")] = request_json.get(
            MCP_METHOD
        )
    if request_json.get(MCP_PARAMS) and request_json.get(MCP_PARAMS).get(
        MCP_CLIENT_INFO
    ):
        session.metadata["mcp_client_name"] = (
            request_json.get(MCP_PARAMS).get(MCP_CLIENT_INFO).get("name", "")
        )

    if request_json.get(MCP_METHOD) == MCP_TOOL_CALL:
        # Intercept and potentially block the request
        hook_tool_call_result, is_blocked = await _hook_tool_call(
            session_id=session_id, request_json=request_json
        )
        if is_blocked:
            # Add the error message to the session.
            # The error message is sent back to the client using the SSE stream.
            await session.add_pending_error_message(hook_tool_call_result)
            return Response(content="Accepted", status_code=202)
    elif request_json.get(MCP_METHOD) == MCP_LIST_TOOLS:
        # Intercept and potentially block the request
        hook_tool_call_result, is_blocked = await _hook_tool_call(
            session_id=session_id, request_json={
                "id": request_json.get("id"),
                "method": MCP_LIST_TOOLS,
                "params": {
                    "name": MCP_LIST_TOOLS,
                    "arguments": {}
                },
            }
        )
        if is_blocked:
            # Add the error message to the session.
            # The error message is sent back to the client using the SSE stream.
            await session.add_pending_error_message(hook_tool_call_result)
            return Response(content="Accepted", status_code=202)

    async with httpx.AsyncClient(timeout=CLIENT_TIMEOUT) as client:
        try:
            response = await client.post(
                url=mcp_server_messages_endpoint,
                headers={
                    k: v
                    for k, v in request.headers.items()
                    if k.lower() in MCP_SERVER_POST_HEADERS
                },
                json=request_json,
                params=query_params,
            )
            return Response(
                content=response.content,
                status_code=response.status_code,
                headers={
                    "X-Proxied-By": "mcp-gateway",
                    **response.headers,
                },
            )

        except httpx.RequestError as e:
            print(f"[MCP POST] Request error: {str(e)}")
            raise HTTPException(status_code=500, detail="Request error") from e
        except Exception as e:
            print(f"[MCP POST] Unexpected error: {str(e)}")
            raise HTTPException(status_code=500, detail="Unexpected error") from e



@gateway.get("/mcp/sse")
async def mcp_get_sse_gateway(
    request: Request,
) -> StreamingResponse:
    """Proxy calls to the MCP Server tools"""
    mcp_server_base_url = request.headers.get("mcp-server-base-url")
    if not mcp_server_base_url:
        print("missing base url", request.headers, flush=True)
        raise HTTPException(
            status_code=400,
            detail="Missing 'mcp-server-base-url' header",
        )
    mcp_server_sse_endpoint = (
        _convert_localhost_to_docker_host(mcp_server_base_url) + "/sse"
    )

    query_params = dict(request.query_params)
    response_headers = {}
    filtered_headers = {
        k: v for k, v in request.headers.items() if k.lower() in MCP_SERVER_SSE_HEADERS
    }
    sse_header_attributes = SseHeaderAttributes.from_request_headers(request.headers)

    async def event_generator():
        """
        Generate a merged stream of MCP server events and pending error messages.
        The pending error messages are added in the POST messages handler.
        This function runs in a loop, yielding events as they arrive.
        """
        mcp_server_events_queue = asyncio.Queue()
        pending_error_messages_queue = asyncio.Queue()
        tasks = set()
        session_id = None

        try:
            # MCP Server Events Processor
            async def process_mcp_server_events():
                """Connect to MCP server and process its events."""
                nonlocal session_id

                async with httpx.AsyncClient(
                    timeout=httpx.Timeout(CLIENT_TIMEOUT)
                ) as client:
                    try:
                        async with aconnect_sse(
                            client,
                            "GET",
                            mcp_server_sse_endpoint,
                            headers=filtered_headers,
                            params=query_params,
                        ) as event_source:
                            if event_source.response.status_code != 200:
                                error_content = await event_source.response.aread()
                                raise HTTPException(
                                    status_code=event_source.response.status_code,
                                    detail=error_content,
                                )

                            async for sse in event_source.aiter_sse():
                                if sse.event == "endpoint":
                                    (
                                        event_bytes,
                                        extracted_id,
                                    ) = await _handle_endpoint_event(
                                        sse, sse_header_attributes
                                    )
                                    session_id = extracted_id

                                    if (
                                        session_id
                                        and "process_error_messages_task"
                                        not in locals()
                                    ):
                                        process_error_messages_task = (
                                            asyncio.create_task(
                                                _check_for_pending_error_messages(
                                                    session_id,
                                                    pending_error_messages_queue,
                                                )
                                            )
                                        )
                                        tasks.add(process_error_messages_task)
                                        process_error_messages_task.add_done_callback(
                                            tasks.discard
                                        )

                                elif sse.event == "message" and session_id:
                                    # Process message event
                                    event_bytes = await _handle_message_event(
                                        session_id, sse
                                    )
                                else:
                                    # Pass through other event types
                                    # pylint: disable=line-too-long
                                    event_bytes = f"event: {sse.event}\ndata: {sse.data}\n\n".encode(
                                        "utf-8"
                                    )

                                # Put the processed event in the queue
                                await mcp_server_events_queue.put(event_bytes)

                    except httpx.StreamClosed as e:
                        print(f"Server stream closed: {e}", flush=True)
                    except Exception as e:
                        print(f"Error processing server events: {e}", flush=True)

            # Start server events processor
            mcp_server_events_task = asyncio.create_task(process_mcp_server_events())
            tasks.add(mcp_server_events_task)
            mcp_server_events_task.add_done_callback(tasks.discard)

            # Main event loop: merge MCP server events and pending error messages
            while True:
                # Create futures for both queues
                mcp_server_event_future = asyncio.create_task(
                    mcp_server_events_queue.get()
                )
                pending_error_message_future = asyncio.create_task(
                    pending_error_messages_queue.get()
                )

                # Wait for either queue to have an item, with timeout
                done, pending = await asyncio.wait(
                    [mcp_server_event_future, pending_error_message_future],
                    return_when=asyncio.FIRST_COMPLETED,
                    timeout=0.25,
                )

                for future in pending:
                    future.cancel()

                # Timeout occurred and no future completed.
                if not done:
                    continue

                for future in done:
                    try:
                        event = await future
                        yield event
                    except asyncio.CancelledError:
                        # Future was cancelled, continue
                        continue

        finally:
            # Clean up all tasks
            for task in tasks:
                task.cancel()

            # Wait for all tasks to complete
            if tasks:
                await asyncio.wait(tasks, timeout=2)

    # Return the streaming response
    return StreamingResponse(
        event_generator(),
        media_type="text/event-stream",
        headers={"X-Proxied-By": "mcp-gateway", **response_headers},
    )


async def _hook_tool_call(session_id: str, request_json: dict) -> Tuple[dict, bool]:
    """
    Hook to process the request JSON before sending it to the MCP server.

    Args:
        session_id (str): The session ID associated with the request.
        request_json (dict): The request JSON to be processed.
    """
    tool_call = {
        "id": f"call_{request_json.get('id')}",
        "type": "function",
        "function": {
            "name": request_json.get(MCP_PARAMS).get("name"),
            "arguments": request_json.get(MCP_PARAMS).get("arguments"),
        },
    }
    message = {"role": "assistant", "content": "", "tool_calls": [tool_call]}
    # Check for blocking guardrails - this blocks until completion
    session = session_store.get_session(session_id)
    guardrails_result = await session.get_guardrails_check_result(
        message, action=GuardrailAction.BLOCK
    )
    # If the request is blocked, return a message indicating the block reason.
    # If there are new errors, run append_and_push_trace in background.
    # If there are no new errors, just return the original request.
    if (
        guardrails_result
        and guardrails_result.get("errors", [])
        and _check_if_new_errors(session_id, guardrails_result)
    ):
        # Add the trace to the explorer
        asyncio.create_task(
            session_store.add_message_to_session(
                session_id=session_id,
                message=message,
                guardrails_result=guardrails_result,
            )
        )
        return {
            "jsonrpc": "2.0",
            "id": request_json.get("id"),
            "error": {
                "code": -32600,
                "message": INVARIANT_GUARDRAILS_BLOCKED_MESSAGE
                % guardrails_result["errors"],
            },
        }, True
    # Push trace to the explorer - don't block on its response
    await session_store.add_message_to_session(session_id, message, guardrails_result)
    return request_json, False


async def _hook_tool_call_response(session_id: str, response_json: dict, is_tools_list=False) -> dict:
    """

    Hook to process the response JSON after receiving it from the MCP server.
    Args:
        session_id (str): The session ID associated with the request.
        response_json (dict): The response JSON to be processed.
    Returns:
        dict: The response JSON is returned if no guardrail is violated
              else an error dict is returned.
    """
    blocked = False
    message = {
        "role": "tool",
        "tool_call_id": f"call_{response_json.get('id')}",
        "content": response_json.get(MCP_RESULT).get("content"),
        "error": response_json.get(MCP_RESULT).get("error"),
    }
    result = response_json
    session = session_store.get_session(session_id)
    guardrails_result = await session.get_guardrails_check_result(
        message, action=GuardrailAction.BLOCK
    )

    if (
        guardrails_result
        and guardrails_result.get("errors", [])
        and _check_if_new_errors(session_id, guardrails_result)
    ):
        blocked = True
        # If the request is blocked, return a message indicating the block reason.
        if not is_tools_list:
            result = {
                "jsonrpc": "2.0",
                "id": response_json.get("id"),
                "error": {
                    "code": -32600,
                    "message": INVARIANT_GUARDRAILS_BLOCKED_MESSAGE
                    % guardrails_result["errors"],
                },
            }
        else:
            # special error response for tools/list tool call
            result = {
                    "jsonrpc": "2.0",
                    "id": response_json.get("id"),
                    "result": {
                        "tools": [
                        {
                            "name": "blocked_" + tool["name"],
                            "description": INVARIANT_GUARDRAILS_BLOCKED_TOOLS_MESSAGE
                            % format_errors_in_response(guardrails_result["errors"]),
                            # no parameters
                            "inputSchema": {
                                "properties": {},
                                "required": [],
                                "title": "invariant_mcp_server_blockedArguments",
                                "type": "object",
                            },
                            "annotations": {
                                "title": "This tool was blocked by security guardrails.",
                            },
                        }
                        for tool in response_json["result"]["tools"]
                    ]
                }
            }

    # Push trace to the explorer - don't block on its response
    asyncio.create_task(
        session_store.add_message_to_session(session_id, message, guardrails_result)
    )
    return result, blocked


def _convert_localhost_to_docker_host(mcp_server_base_url: str) -> str:
    """
    Convert localhost or 127.0.0.1 in an address to host.docker.internal

    Args:
        mcp_server_base_url (str): The original server address from the header

    Returns:
        str: Modified server address with localhost references changed to host.docker.internal
    """
    # check if we are running in a docker container
    if not os.environ.get("DOCKER_ENV"):
        return mcp_server_base_url

    if "localhost" in mcp_server_base_url or "127.0.0.1" in mcp_server_base_url:
        # Replace localhost or 127.0.0.1 with host.docker.internal
        modified_address = re.sub(
            r"(https?://)(?:localhost|127\.0\.0\.1)(\b|:)",
            r"\1host.docker.internal\2",
            mcp_server_base_url,
        )
        return modified_address

    return mcp_server_base_url


async def _handle_endpoint_event(
    sse: ServerSentEvent, sse_header_attributes: SseHeaderAttributes
) -> Tuple[bytes, str]:
    """
    Handle the endpoint event type and modify the data accordingly.
    For endpoint events, we need to rewrite the endpoint to use our gateway.

    Args:
        sse (ServerSentEvent): The original SSE object.
        sse_header_attributes (SseHeaderAttributes): The header attributes from the request.

    Returns:
        bytes: Modified SSE data as bytes.
        str: session_id extracted from the data.
    """
    # Extract session_id
    match = re.search(r"session_id=([^&\s]+)", sse.data)
    if match:
        session_id = match.group(1)
        # Initialize this session in our store if needed
        if not session_store.session_exists(session_id):
            await session_store.initialize_session(session_id, sse_header_attributes)

    # Rewrite the endpoint to use our gateway
    modified_data = sse.data.replace(
        "/messages/?session_id=",
        "/api/v1/gateway/mcp/sse/messages/?session_id=",
    )
    event_bytes = f"event: {sse.event}\ndata: {modified_data}\n\n".encode("utf-8")
    return event_bytes, session_id


async def _handle_message_event(session_id: str, sse: ServerSentEvent) -> bytes:
    """
    Handle the message event type.

    Args:
        session_id (str): The session ID associated with the request.
        sse (ServerSentEvent): The original SSE object.
    """
    event_bytes = f"event: {sse.event}\ndata: {sse.data}\n\n".encode("utf-8")
    session = session_store.get_session(session_id)
    try:
        response_json = json.loads(sse.data)

        if response_json.get(MCP_RESULT) and response_json.get(MCP_RESULT).get(
            MCP_SERVER_INFO
        ):
            session.metadata["mcp_server_name"] = (
                response_json.get(MCP_RESULT).get(MCP_SERVER_INFO).get("name", "")
            )

        method = session.id_to_method_mapping.get(response_json.get("id"))
        if method == MCP_TOOL_CALL:
            hook_tool_call_response, blocked = await _hook_tool_call_response(
                session_id=session_id,
                response_json=response_json,
            )
            # Update the event bytes with hook_tool_call_response.
            # hook_tool_call_response is same as response_json if no guardrail is violated.
            # If guardrail is violated, it contains the error message.
            # pylint: disable=line-too-long
            if blocked:
                event_bytes = f"event: {sse.event}\ndata: {json.dumps(hook_tool_call_response)}\n\n".encode(
                    "utf-8"
                )
        elif method == MCP_LIST_TOOLS:
            # store tools in metadata
            session_store.get_session(session_id).metadata["tools"] = response_json.get(
                MCP_RESULT
            ).get("tools")
            # store tools/list tool call in trace
            hook_tool_call_response, blocked = await _hook_tool_call_response(
                session_id=session_id,
                response_json={
                    "id": response_json.get("id"),
                    "result": {
                        "content": json.dumps(response_json.get(MCP_RESULT).get("tools")),
                        "tools": response_json.get(MCP_RESULT).get("tools"),
                    }
                },
                is_tools_list=True,
            )
            # Update the event bytes with hook_tool_call_response.
            # hook_tool_call_response is same as response_json if no guardrail is violated.
            # If guardrail is violated, it contains the error message.
            # pylint: disable=line-too-long
            if blocked:
                event_bytes = f"event: {sse.event}\ndata: {json.dumps(hook_tool_call_response)}\n\n".encode(
                    "utf-8"
                )

    except json.JSONDecodeError as e:
        print(
            f"[MCP SSE] Error parsing message JSON: {e}",
            flush=True,
        )
    except Exception as e:  # pylint: disable=broad-except
        print(
            f"[MCP SSE] Error processing message: {e}",
            flush=True,
        )
    return event_bytes


def _check_if_new_errors(session_id: str, guardrails_result: dict) -> bool:
    """Checks if there are new errors in the guardrails result."""
    session = session_store.get_session(session_id)
    annotations = create_annotations_from_guardrails_errors(
        guardrails_result.get("errors", [])
    )
    for annotation in annotations:
        if annotation not in session.annotations:
            return True
    return False


async def _check_for_pending_error_messages(
    session_id: str, pending_error_messages_queue: asyncio.Queue
):
    """Periodically check for and enqueue pending error messages."""
    try:
        while True:
            try:
                session = session_store.get_session(session_id)
                error_messages = await session.get_pending_error_messages()

                for error_message in error_messages:
                    error_bytes = (
                        f"event: message\ndata: {json.dumps(error_message)}\n\n".encode(
                            "utf-8"
                        )
                    )
                    await pending_error_messages_queue.put(error_bytes)

                await asyncio.sleep(1)
            except Exception as e:  # pylint: disable=broad-except
                print(f"Error checking for messages: {e}", flush=True)
                await asyncio.sleep(1)
    except asyncio.CancelledError:
        # Task was cancelled, exit gracefully
        return



"""Gateway service to forward requests to the Gemini APIs"""

import asyncio
import json
from typing import Any, Literal, Optional

import httpx
from fastapi import APIRouter, Depends, HTTPException, Query, Request, Response
from fastapi.responses import StreamingResponse

from gateway.common.authorization import extract_authorization_from_headers
from gateway.common.config_manager import (
    GatewayConfig,
    GatewayConfigManager,
    extract_guardrails_from_header,
)
from gateway.common.constants import (
    CLIENT_TIMEOUT,
    IGNORED_HEADERS,
)
from gateway.common.guardrails import GuardrailAction, GuardrailRuleSet
from gateway.common.request_context import RequestContext
from gateway.converters.gemini_to_invariant import convert_request, convert_response
from gateway.integrations.explorer import (
    create_annotations_from_guardrails_errors,
    fetch_guardrails_from_explorer,
    push_trace,
)
from gateway.integrations.guardrails import (
    ExtraItem,
    InstrumentedResponse,
    InstrumentedStreamingResponse,
    Replacement,
    check_guardrails,
)

gateway = APIRouter()

GEMINI_AUTHORIZATION_HEADER = "x-goog-api-key"
GEMINI_AUTHORIZATION_FALLBACK_HEADER = "authorization"


@gateway.post("/gemini/{api_version}/models/{model}:{endpoint}")
@gateway.post("/{dataset_name}/gemini/{api_version}/models/{model}:{endpoint}")
async def gemini_generate_content_gateway(
    request: Request,
    api_version: str,
    model: str,
    endpoint: str,
    dataset_name: str = None,  # This is None if the client doesn't want to push to Explorer
    alt: str = Query(
        None, title="Response Format", description="Set to 'sse' for streaming"
    ),
    config: GatewayConfig = Depends(GatewayConfigManager.get_config),  # pylint: disable=unused-argument
    header_guardrails: GuardrailRuleSet = Depends(extract_guardrails_from_header),
) -> Response:
    """Proxy calls to the Gemini GenerateContent API"""
    if endpoint not in ["generateContent", "streamGenerateContent"]:
        return Response(
            content="Invalid endpoint - the only endpoints supported are: \
            /api/v1/gateway/gemini/<version>/models/<model-name>:generateContent or \
            /api/v1/gateway/<dataset-name>/gemini/<version>models/<model-name>:generateContent",
            status_code=400,
        )
    headers = {
        k: v
        for k, v in request.headers.items()
        if k.lower() not in IGNORED_HEADERS + [GEMINI_AUTHORIZATION_FALLBACK_HEADER]
    }
    headers["accept-encoding"] = "identity"
    invariant_authorization, gemini_api_key = extract_authorization_from_headers(
        request,
        dataset_name,
        GEMINI_AUTHORIZATION_HEADER,
        [GEMINI_AUTHORIZATION_FALLBACK_HEADER],
    )
    headers[GEMINI_AUTHORIZATION_HEADER] = gemini_api_key

    request_body_bytes = await request.body()
    request_json = json.loads(request_body_bytes)

    client = httpx.AsyncClient(timeout=httpx.Timeout(CLIENT_TIMEOUT))
    gemini_api_url = f"https://generativelanguage.googleapis.com/{api_version}/models/{model}:{endpoint}"
    if alt == "sse":
        gemini_api_url += "?alt=sse"
    gemini_request = client.build_request(
        "POST",
        gemini_api_url,
        content=request_body_bytes,
        headers=headers,
    )

    dataset_guardrails = None
    if dataset_name:
        # Get the guardrails for the dataset
        dataset_guardrails = await fetch_guardrails_from_explorer(
            dataset_name, invariant_authorization
        )
    context = RequestContext.create(
        request_json=request_json,
        dataset_name=dataset_name,
        invariant_authorization=invariant_authorization,
        guardrails=header_guardrails or dataset_guardrails,
        config=config,
        request=request,
    )
    if alt == "sse" or endpoint == "streamGenerateContent":
        return await stream_response(
            context,
            client,
            gemini_request,
        )
    return await handle_non_streaming_response(
        context,
        client,
        gemini_request,
    )


class InstrumentedStreamingGeminiResponse(InstrumentedStreamingResponse):
    """Instrumented streaming response for Gemini API"""

    def __init__(
        self,
        context: RequestContext,
        client: httpx.AsyncClient,
        gemini_request: httpx.Request,
    ):
        super().__init__()

        # request data
        self.context: RequestContext = context
        self.client: httpx.AsyncClient = client
        self.gemini_request: httpx.Request = gemini_request

        # Store the progressively merged response
        self.merged_response = {
            "candidates": [{"content": {"parts": []}, "finishReason": None}]
        }

        # guardrailing execution result (if any)
        self.guardrails_execution_result: Optional[dict[str, Any]] = None

    def make_refusal(
        self,
        location: Literal["request", "response"],
        guardrails_execution_result: dict[str, Any],
    ) -> dict:
        """Create a refusal response for the given request or response"""
        return {
            "candidates": [
                {
                    "content": {
                        "parts": [
                            {
                                "text": f"[Invariant] The {location} did not pass the guardrails",
                            }
                        ],
                    }
                }
            ],
            "error": {
                "code": 400,
                "message": f"[Invariant] The {location} did not pass the guardrails",
                "details": guardrails_execution_result,
                "status": "INVARIANT_GUARDRAILS_VIOLATION",
            },
            "promptFeedback": {
                "blockReason": "SAFETY",
                "block_reason_message": f"[Invariant] The {location} did not pass the guardrails: "
                + json.dumps(guardrails_execution_result),
                "safetyRatings": [
                    {
                        "category": "HARM_CATEGORY_UNSPECIFIED",
                        "probability": "HIGH",
                        "blocked": True,
                    }
                ],
            },
        }

    async def on_start(self):
        """
        Check guardrails in a pipelined fashion, before processing the first chunk
        (for input guardrailing).
        """
        if self.context.guardrails:
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context, action=GuardrailAction.BLOCK, response_json={}
            )
            if self.guardrails_execution_result.get("errors", []):
                error_chunk = json.dumps(
                    self.make_refusal("request", self.guardrails_execution_result)
                )

                # Push annotated trace to the explorer - don't block on its response
                if self.context.dataset_name:
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            {},
                            self.guardrails_execution_result,
                        )
                    )

                # if we find something, we end the stream prematurely (end_of_stream=True)
                # and yield an error chunk instead of actually beginning the stream
                return ExtraItem(
                    f"data: {error_chunk}\r\n\r\n".encode(), end_of_stream=True
                )

    async def event_generator(self):
        """Event generator for streaming responses"""
        response = await self.client.send(self.gemini_request, stream=True)

        if response.status_code != 200:
            error_content = await response.aread()
            try:
                error_json = json.loads(error_content.decode("utf-8"))
                error_detail = error_json.get("error", "Unknown error from Gemini API")
            except json.JSONDecodeError:
                error_detail = {"error": "Failed to parse Gemini error response"}
            raise HTTPException(status_code=response.status_code, detail=error_detail)

        async for chunk in response.aiter_bytes():
            yield chunk

    async def on_chunk(self, chunk):
        """Processes each chunk of the streaming response"""
        chunk_text = chunk.decode().strip()
        if not chunk_text:
            return

        # Parse and update merged_response incrementally
        process_chunk_text(self.merged_response, chunk_text)

        # runs on the last stream item
        if (
            self.merged_response.get("candidates", [])
            and self.merged_response.get("candidates")[0].get("finishReason", "")
            and self.context.guardrails
        ):
            # Block on the guardrails check
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context,
                action=GuardrailAction.BLOCK,
                response_json=self.merged_response,
            )
            if self.guardrails_execution_result.get("errors", []):
                error_chunk = json.dumps(
                    self.make_refusal("response", self.guardrails_execution_result)
                )

                # Push annotated trace to the explorer - don't block on its response
                if self.context.dataset_name:
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            self.merged_response,
                            self.guardrails_execution_result,
                        )
                    )

                return ExtraItem(
                    value=f"data: {error_chunk}\r\n\r\n".encode(),
                    # for Gemini we have to end the stream prematurely, as the client SDK
                    # will not stop streaming when it encounters an error
                    end_of_stream=True,
                )

    async def on_end(self):
        """Runs when the stream ends."""

        # Push annotated trace to the explorer - don't block on its response
        if self.context.dataset_name:
            asyncio.create_task(
                push_to_explorer(
                    self.context,
                    self.merged_response,
                    self.guardrails_execution_result,
                )
            )


async def stream_response(
    context: RequestContext,
    client: httpx.AsyncClient,
    gemini_request: httpx.Request,
) -> Response:
    """Handles streaming the Gemini response to the client"""

    response = InstrumentedStreamingGeminiResponse(
        context=context,
        client=client,
        gemini_request=gemini_request,
    )

    async def event_generator():
        async for chunk in response.instrumented_event_generator():
            yield chunk

    return StreamingResponse(
        event_generator(),
        media_type="text/event-stream",
    )


def process_chunk_text(
    merged_response: dict[str, Any],
    chunk_text: str,
) -> None:
    """Processes the chunk text and updates the merged_response to be sent to the explorer"""
    # Split the chunk text into individual JSON strings
    # A single chunk can contain multiple "data: " sections
    for json_string in chunk_text.split("data: "):
        json_string = json_string.replace("data: ", "").strip()

        if not json_string:
            continue

        try:
            json_chunk = json.loads(json_string)
        except json.JSONDecodeError:
            print("Warning: Could not parse chunk:", json_string)

        update_merged_response(merged_response, json_chunk)


def update_merged_response(merged_response: dict[str, Any], chunk_json: dict) -> None:
    """Updates the merged response incrementally with a new chunk."""
    candidates = chunk_json.get("candidates", [])

    for candidate in candidates:
        content = candidate.get("content", {})
        parts = content.get("parts", [])

        for part in parts:
            if "text" in part:
                existing_parts = merged_response["candidates"][0]["content"]["parts"]
                if existing_parts and "text" in existing_parts[-1]:
                    existing_parts[-1]["text"] += part["text"]
                else:
                    existing_parts.append({"text": part["text"]})

            if "functionCall" in part:
                merged_response["candidates"][0]["content"]["parts"].append(
                    {"functionCall": part["functionCall"]}
                )

        if "role" in content:
            merged_response["candidates"][0]["content"]["role"] = content["role"]

        if "finishReason" in candidate:
            merged_response["candidates"][0]["finishReason"] = candidate["finishReason"]

    if "usageMetadata" in chunk_json:
        merged_response["usageMetadata"] = chunk_json["usageMetadata"]
    if "modelVersion" in chunk_json:
        merged_response["modelVersion"] = chunk_json["modelVersion"]


def create_metadata(
    context: RequestContext, response_json: dict[str, Any]
) -> dict[str, Any]:
    """Creates metadata for the trace"""
    metadata = {
        k: v
        for k, v in context.request_json.items()
        if k not in ("systemInstruction", "contents")
    }
    metadata["via_gateway"] = True
    metadata.update(
        {
            key: value
            for key, value in response_json.items()
            if key in ("usageMetadata", "modelVersion")
        }
    )
    return metadata


async def get_guardrails_check_result(
    context: RequestContext, action: GuardrailAction, response_json: dict[str, Any]
) -> dict[str, Any]:
    """Get the guardrails check result"""
    # Determine which guardrails to apply based on the action
    guardrails = (
        context.guardrails.logging_guardrails
        if action == GuardrailAction.LOG
        else context.guardrails.blocking_guardrails
    )
    if not guardrails:
        return {}

    converted_requests = convert_request(context.request_json)
    converted_responses = convert_response(response_json)

    # Block on the guardrails check
    guardrails_execution_result = await check_guardrails(
        messages=converted_requests + converted_responses,
        guardrails=guardrails,
        context=context,
    )
    return guardrails_execution_result


async def push_to_explorer(
    context: RequestContext,
    response_json: dict[str, Any],
    guardrails_execution_result: Optional[dict] = None,
) -> None:
    """Pushes the full trace to the Invariant Explorer"""
    guardrails_execution_result = guardrails_execution_result or {}
    annotations = create_annotations_from_guardrails_errors(
        guardrails_execution_result.get("errors", [])
    )

    # Execute the logging guardrails before pushing to Explorer
    logging_guardrails_execution_result = await get_guardrails_check_result(
        context,
        action=GuardrailAction.LOG,
        response_json=response_json,
    )
    logging_annotations = create_annotations_from_guardrails_errors(
        logging_guardrails_execution_result.get("errors", [])
    )
    # Update the annotations with the logging guardrails
    annotations.extend(logging_annotations)

    converted_requests = convert_request(context.request_json)
    converted_responses = convert_response(response_json)

    _ = await push_trace(
        dataset_name=context.dataset_name,
        messages=[converted_requests + converted_responses],
        invariant_authorization=context.invariant_authorization,
        metadata=[create_metadata(context, response_json)],
        annotations=[annotations] if annotations else None,
    )


class InstrumentedGeminiResponse(InstrumentedResponse):
    """Instrumented response for Gemini API"""

    def __init__(
        self,
        context: RequestContext,
        client: httpx.AsyncClient,
        gemini_request: httpx.Request,
    ):
        super().__init__()

        # request data
        self.context: RequestContext = context
        self.client: httpx.AsyncClient = client
        self.gemini_request: httpx.Request = gemini_request

        # response data
        self.response: Optional[httpx.Response] = None
        self.response_json: Optional[dict[str, Any]] = None

        # guardrails execution result (if any)
        self.guardrails_execution_result: Optional[dict[str, Any]] = None

    async def on_start(self):
        """
        Check guardrails in a pipelined fashion, before processing the first chunk
        (for input guardrailing).
        """
        if self.context.guardrails:
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context, action=GuardrailAction.BLOCK, response_json={}
            )
            if self.guardrails_execution_result.get("errors", []):
                error_chunk = json.dumps(
                    {
                        "error": {
                            "code": 400,
                            "message": "[Invariant] The request did not pass the guardrails",
                            "details": self.guardrails_execution_result,
                            "status": "INVARIANT_GUARDRAILS_VIOLATION",
                        },
                        "prompt_feedback": {
                            "blockReason": "SAFETY",
                            "safetyRatings": [
                                {
                                    "category": "HARM_CATEGORY_UNSPECIFIED",
                                    "probability": 0.0,
                                    "blocked": True,
                                }
                            ],
                        },
                    }
                )

                # Push annotated trace to the explorer - don't block on its response
                if self.context.dataset_name:
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            {},
                            self.guardrails_execution_result,
                        )
                    )

                # if we find something, we end the stream prematurely (end_of_stream=True)
                # and yield an error chunk instead of actually beginning the stream
                return Replacement(
                    Response(
                        content=error_chunk,
                        status_code=400,
                        media_type="application/json",
                        headers={
                            "Content-Type": "application/json",
                        },
                    )
                )

    async def request(self):
        """Makes the request to the Gemini API and return the response"""
        self.response = await self.client.send(self.gemini_request)

        response_string = self.response.text
        response_code = self.response.status_code

        try:
            self.response_json = self.response.json()
        except json.JSONDecodeError as e:
            raise HTTPException(
                status_code=self.response.status_code,
                detail="Invalid JSON response received from Gemini API",
            ) from e
        if self.response.status_code != 200:
            raise HTTPException(
                status_code=self.response.status_code,
                detail=self.response_json.get("error", "Unknown error from Gemini API"),
            )

        return Response(
            content=response_string,
            status_code=response_code,
            media_type="application/json",
            headers=dict(self.response.headers),
        )

    async def on_end(self):
        """Runs when the request ends."""
        response_string = json.dumps(self.response_json)
        response_code = self.response.status_code

        if self.context.guardrails:
            # Block on the guardrails check
            guardrails_execution_result = await get_guardrails_check_result(
                self.context,
                action=GuardrailAction.BLOCK,
                response_json=self.response_json,
            )
            if guardrails_execution_result.get("errors", []):
                response_string = json.dumps(
                    {
                        "error": {
                            "code": 400,
                            "message": "[Invariant] The response did not pass the guardrails",
                            "details": guardrails_execution_result,
                            "status": "INVARIANT_GUARDRAILS_VIOLATION",
                        },
                    }
                )
                response_code = 400

                if self.context.dataset_name:
                    # Push to Explorer - don't block on its response
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            self.response_json,
                            guardrails_execution_result,
                        )
                    )

                return Replacement(
                    Response(
                        content=response_string,
                        status_code=response_code,
                        media_type="application/json",
                        headers=dict(self.response.headers),
                    )
                )

        # Otherwise, also push to Explorer - don't block on its response
        if self.context.dataset_name:
            asyncio.create_task(
                push_to_explorer(
                    self.context, self.response_json, guardrails_execution_result
                )
            )


async def handle_non_streaming_response(
    context: RequestContext,
    client: httpx.AsyncClient,
    gemini_request: httpx.Request,
) -> Response:
    """Handles non-streaming Gemini responses"""

    response = InstrumentedGeminiResponse(
        context=context,
        client=client,
        gemini_request=gemini_request,
    )

    return await response.instrumented_request()



"""Validates the GatewayConfigManager configuration."""

import sys

from gateway.common.config_manager import GatewayConfigManager

try:
    _ = GatewayConfigManager.get_config()
    print("[gateway config validated successfully]")
    sys.exit(0)
except Exception as e:
    print(f"Error loading GatewayConfig: {e}")
    sys.exit(1)



"""MCP Sessions Manager related classes"""

import asyncio
import contextlib
import os
import random

from typing import Any, Dict, List, Optional

from invariant_sdk.async_client import AsyncClient
from invariant_sdk.types.append_messages import AppendMessagesRequest
from invariant_sdk.types.push_traces import PushTracesRequest
from pydantic import BaseModel, Field, PrivateAttr
from starlette.datastructures import Headers

from gateway.common.guardrails import GuardrailRuleSet, GuardrailAction
from gateway.common.request_context import RequestContext
from gateway.integrations.explorer import (
    create_annotations_from_guardrails_errors,
    fetch_guardrails_from_explorer,
)
from gateway.integrations.guardrails import check_guardrails

DEFAULT_API_URL = "https://explorer.invariantlabs.ai"


class McpSession(BaseModel):
    """
    Represents a single MCP session.
    """

    session_id: str
    messages: List[Dict[str, Any]] = Field(default_factory=list)
    metadata: Dict[str, Any] = Field(default_factory=dict)
    id_to_method_mapping: Dict[int, str] = Field(default_factory=dict)
    explorer_dataset: str
    push_explorer: bool
    trace_id: Optional[str] = None
    last_trace_length: int = 0
    annotations: List[Dict[str, Any]] = Field(default_factory=list)
    guardrails: GuardrailRuleSet = Field(
        default_factory=lambda: GuardrailRuleSet(
            blocking_guardrails=[], logging_guardrails=[]
        )
    )
    # When tool calls are blocked, the error message is stored here
    # and sent to the client via the SSE stream.
    pending_error_messages: List[dict] = Field(default_factory=list)

    # Lock to maintain in-order pushes to explorer
    _lock: asyncio.Lock = PrivateAttr(default_factory=asyncio.Lock)

    async def load_guardrails(self) -> None:
        """
        Load guardrails for the session.

        This method fetches guardrails from the Invariant Explorer and assigns them to the session.
        """
        self.guardrails = await fetch_guardrails_from_explorer(
            self.explorer_dataset,
            "Bearer " + os.getenv("INVARIANT_API_KEY"),
        )

    def _deduplicate_annotations(self, new_annotations: list) -> list:
        """Deduplicate new_annotations using the annotations in the session."""
        deduped_annotations = []
        for annotation in new_annotations:
            # Check if an annotation with the same content and address exists in self.annotations
            # TODO: Rely on the __eq__ method of the AnnotationCreate class directly via not in
            # to remove duplicates instead of using a custom logic.
            # This is a temporary solution until the Invariant SDK is updated.
            is_duplicate = False
            for current_annotation in self.annotations:
                if (
                    annotation.content == current_annotation.content
                    and annotation.address == current_annotation.address
                    and annotation.extra_metadata == current_annotation.extra_metadata
                ):
                    is_duplicate = True
                    break

            if not is_duplicate:
                deduped_annotations.append(annotation)

        return deduped_annotations

    @contextlib.asynccontextmanager
    async def session_lock(self):
        """
        Context manager for the session lock.

        Usage:
        async with session.session_lock():
            # Code that requires exclusive access to the session
        """
        async with self._lock:
            yield

    async def get_guardrails_check_result(
        self,
        message: dict,
        action: GuardrailAction = GuardrailAction.BLOCK,
    ) -> dict:
        """
        Check against guardrails of type action.
        """
        # Skip if no guardrails are configured for this action
        if not (
            (self.guardrails.blocking_guardrails and action == GuardrailAction.BLOCK)
            or (self.guardrails.logging_guardrails and action == GuardrailAction.LOG)
        ):
            return {}

        # Prepare context and select appropriate guardrails
        context = RequestContext.create(
            request_json={},
            dataset_name=self.explorer_dataset,
            invariant_authorization="Bearer " + os.getenv("INVARIANT_API_KEY"),
            guardrails=self.guardrails,
        )

        guardrails_to_check = (
            self.guardrails.blocking_guardrails
            if action == GuardrailAction.BLOCK
            else self.guardrails.logging_guardrails
        )

        result = await check_guardrails(
            messages=self.messages + [message],
            guardrails=guardrails_to_check,
            context=context,
        )
        return result

    async def add_message(
        self, message: Dict[str, Any], guardrails_result=Dict
    ) -> None:
        """
        Add a message to the session and optionally push to explorer.

        Args:
            message: The message to add
        """
        async with self.session_lock():
            annotations = []
            if guardrails_result and guardrails_result.get("errors", []):
                annotations = create_annotations_from_guardrails_errors(
                    guardrails_result.get("errors")
                )

            if self.guardrails.logging_guardrails:
                logging_guardrails_check_result = (
                    await self.get_guardrails_check_result(
                        message, action=GuardrailAction.LOG
                    )
                )
                if (
                    logging_guardrails_check_result
                    and logging_guardrails_check_result.get("errors", [])
                ):
                    annotations.extend(
                        create_annotations_from_guardrails_errors(
                            logging_guardrails_check_result["errors"]
                        )
                    )
            deduplicated_annotations = self._deduplicate_annotations(annotations)
            # pylint: disable=no-member
            self.messages.append(message)
            # If push_explorer is enabled, push the trace
            if self.push_explorer:
                await self._push_trace_update(deduplicated_annotations)

    async def _push_trace_update(self, deduplicated_annotations: list) -> None:
        """
        Push trace updates to the explorer.

        If a trace doesn't exist, create a new one. If it does, append new messages.

        This is an internal method that should only be called within a lock.
        """
        try:
            client = AsyncClient(
                api_url=os.getenv("INVARIANT_API_URL", DEFAULT_API_URL),
            )

            # If no trace exists, create a new one
            if not self.trace_id:
                # pylint: disable=no-member
                metadata = {"source": "mcp", "tools": self.metadata.get("tools", [])}
                if self.metadata.get("mcp_client_name"):
                    metadata["mcp_client"] = self.metadata.get("mcp_client_name")
                if self.metadata.get("mcp_server_name"):
                    metadata["mcp_server"] = self.metadata.get("mcp_server_name")

                response = await client.push_trace(
                    PushTracesRequest(
                        messages=[self.messages],
                        dataset=self.explorer_dataset,
                        metadata=[metadata],
                        annotations=[deduplicated_annotations],
                    )
                )
                self.trace_id = response.id[0]
            else:
                new_messages = self.messages[self.last_trace_length :]
                if new_messages:
                    await client.append_messages(
                        AppendMessagesRequest(
                            trace_id=self.trace_id,
                            messages=new_messages,
                            annotations=deduplicated_annotations,
                        )
                    )
            # pylint: disable=no-member
            self.annotations.extend(deduplicated_annotations)
            self.last_trace_length = len(self.messages)
        except Exception as e:  # pylint: disable=broad-except
            print(f"[MCP SSE] Error pushing trace for session {self.session_id}: {e}")

    async def add_pending_error_message(self, error_message: dict) -> None:
        """
        Add a pending error message to the session.

        Args:
            error_message: The error message to add
        """
        async with self.session_lock():
            # pylint: disable=no-member
            self.pending_error_messages.append(error_message)

    async def get_pending_error_messages(self) -> List[dict]:
        """
        Get all pending error messages for the session.

        Returns:
            List[dict]: A list of pending error messages
        """
        async with self.session_lock():
            messages = list(self.pending_error_messages)
            self.pending_error_messages = []
            return messages


class SseHeaderAttributes(BaseModel):
    """
    A Pydantic model to represent header attributes.
    """

    push_explorer: bool
    explorer_dataset: str

    @classmethod
    def from_request_headers(cls, headers: Headers) -> "SseHeaderAttributes":
        """
        Create an instance from FastAPI request headers.

        Args:
            headers: FastAPI Request headers

        Returns:
            SseHeaderAttributes: An instance with values extracted from headers
        """
        # Extract and process header values
        project_name = headers.get("INVARIANT-PROJECT-NAME")
        push_explorer_header = headers.get("PUSH-INVARIANT-EXPLORER", "false").lower()

        # Determine explorer_dataset
        if project_name:
            explorer_dataset = project_name
        else:
            explorer_dataset = f"mcp-capture-{random.randint(1, 100)}"

        # Determine push_explorer
        push_explorer = push_explorer_header == "true"

        # Create and return instance
        return cls(push_explorer=push_explorer, explorer_dataset=explorer_dataset)


class McpSessionsManager:
    """
    A class to manage MCP sessions and their messages.
    """

    def __init__(self):
        self._sessions: dict[str, McpSession] = {}

    def session_exists(self, session_id: str) -> bool:
        """Check if a session exists"""
        return session_id in self._sessions

    async def initialize_session(
        self, session_id: str, sse_header_attributes: SseHeaderAttributes
    ) -> None:
        """Initialize a new session"""
        if session_id not in self._sessions:
            session = McpSession(
                session_id=session_id,
                explorer_dataset=sse_header_attributes.explorer_dataset,
                push_explorer=sse_header_attributes.push_explorer,
            )
            self._sessions[session_id] = session
            # Load guardrails for the session from the explorer
            await session.load_guardrails()

    def get_session(self, session_id: str) -> McpSession:
        """Get a session by ID"""
        if session_id not in self._sessions:
            raise ValueError(f"Session {session_id} does not exist.")
        return self._sessions.get(session_id)

    async def add_message_to_session(
        self, session_id: str, message: Dict[str, Any], guardrails_result: dict
    ) -> None:
        """
        Add a message to a session and push to explorer if enabled.

        Args:
            session_id: The session ID
            message: The message to add
            guardrails_result: The result of the guardrails check
        """
        session = self.get_session(session_id)
        await session.add_message(message, guardrails_result)



"""Script is used to run actions using the Invariant Gateway."""

import asyncio
import os
import signal
import subprocess
import sys
import time

from typing import Optional

from gateway.mcp import mcp
from gateway.mcp.log import mcp_log


LOCAL_COMPOSE_FILE = "gateway/docker-compose.local.yml"


# Handle signals to ensure clean shutdown
def signal_handler(sig, frame):
    """Handle signals for graceful shutdown."""
    sys.exit(0)


def print_help():
    """Prints the help message."""
    actions = {
        "mcp": """
        Runs the Invariant Gateway against MCP (Model Context Protocol) stdio servers with guardrailing and push to Explorer features.
        """,
        "server": """
        Runs the Invariant Gateway server locally providing guardrailing and push to Explorer features.
        Should be called with one of the following subcommands: build, up, down, logs.
        A guardrails file can be passed with the flag: --guardrails-file=/path/to/guardrails/file.
        """,
        "help": "Shows this help message.",
    }

    for verb, description in actions.items():
        print(f"{verb}: {description}")


def ensure_network_exists(network_name: str = "invariant-explorer-web") -> bool:
    """Ensure the Docker network exists."""
    try:
        # Check if network exists
        result = subprocess.run(
            ["docker", "network", "inspect", network_name],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False,
        )

        if result.returncode != 0:
            print(f"Creating Docker network: {network_name}")
            subprocess.run(
                ["docker", "network", "create", network_name],
                check=True,
            )

        return True
    except subprocess.CalledProcessError as e:
        print(f"Error creating Docker network: {e}")
        return False


def setup_guardrails(guardrails_file_path: Optional[str] = None) -> bool:
    """Configure guardrails if specified."""
    if not guardrails_file_path:
        return True

    if not os.path.isfile(guardrails_file_path):
        print(
            f"Error: Specified guardrails file does not exist: {guardrails_file_path}"
        )
        return False

    # Convert to absolute path
    guardrails_file_path = os.path.realpath(guardrails_file_path)
    os.environ["GUARDRAILS_FILE_PATH"] = guardrails_file_path

    # Check if INVARIANT_API_KEY is set
    if not os.environ.get("INVARIANT_API_KEY"):
        print(
            "Error: A guardrails file is specified, but INVARIANT_API_KEY env var is not set. "
            "This is required to validate guardrails."
        )
        return False

    return True


def build():
    """Build Docker containers using docker-compose."""
    try:
        print(f"Building using docker-compose file: {LOCAL_COMPOSE_FILE}")
        subprocess.run(
            ["docker", "compose", "-f", str(LOCAL_COMPOSE_FILE), "build"],
            check=True,
        )
        print("Build completed successfully")
        return True
    except subprocess.CalledProcessError as e:
        print(f"Error building containers: {e}")
        return False


def up(guardrails_file_path: Optional[str] = None):
    """Set up the local server for the Invariant Gateway."""
    # Ensure network exists
    if not ensure_network_exists():
        return 1

    # Set up guardrails
    if not setup_guardrails(guardrails_file_path=guardrails_file_path):
        return 1

    # Run the server
    try:
        # Start containers
        print(f"Starting containers using docker-compose file: {LOCAL_COMPOSE_FILE}")
        subprocess.run(
            ["docker", "compose", "-f", str(LOCAL_COMPOSE_FILE), "up", "-d"],
            check=True,
        )

        # Wait for containers to start
        time.sleep(2)

        # Check if gateway container is running
        result = subprocess.run(
            ["docker", "ps", "-qf", "name=invariant-gateway"],
            capture_output=True,
            text=True,
            check=True,
        )

        if not result.stdout.strip():
            print("The invariant-gateway container failed to start.")
            logs = subprocess.run(
                ["docker", "logs", "invariant-gateway"],
                capture_output=True,
                text=True,
                check=False,
            )
            print("Last 20 lines of logs:")
            print("\n".join(logs.stdout.strip().split("\n")[-20:]))
            return False

        print("Gateway started at http://localhost:8005/api/v1/gateway/")
        print("See http://localhost:8005/api/v1/gateway/docs for API documentation")

        if guardrails_file_path:
            print(f"Using Guardrails File: {guardrails_file_path}")

        return True
    except subprocess.CalledProcessError as e:
        print(f"Error starting containers: {e}")
        return False


def down():
    """Stop the Docker containers."""
    try:
        print(f"Stopping containers using docker-compose file: {LOCAL_COMPOSE_FILE}")
        subprocess.run(
            ["docker", "compose", "-f", str(LOCAL_COMPOSE_FILE), "down"],
            check=True,
        )
        print("Containers stopped successfully")
        return True
    except subprocess.CalledProcessError as e:
        print(f"Error stopping containers: {e}")
        return False


def logs():
    """Show container logs."""
    try:
        subprocess.run(
            ["docker", "compose", "-f", str(LOCAL_COMPOSE_FILE), "logs", "-f"],
            check=True,
        )
        return True
    except subprocess.CalledProcessError as e:
        print(f"Error showing logs: {e}")
        return False
    except KeyboardInterrupt:
        print("\nExiting logs view")
        return True


def run_server_command(command, args=None):
    """Run a server command."""
    if args is None:
        args = []

    if command == "build":
        return build()
    elif command == "up":
        # Parse guardrails file from args
        guardrails_file = None
        for arg in args:
            if arg.startswith("--guardrails-file="):
                guardrails_file = arg.split("=", 1)[1]

        return up(guardrails_file)
    elif command == "down":
        return down()
    elif command == "logs":
        return logs()
    else:
        print(f"Unknown server command: {command}")
        print("Available commands: build, up, down, logs")
        return False


def main():
    """Entry point for the Invariant Gateway."""
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    if len(sys.argv) < 2:
        print_help()
        sys.exit(1)

    verb = sys.argv[1]
    if verb == "help":
        print_help()
        return 0

    if "INVARIANT_API_KEY" not in os.environ:
        print("[ERROR] INVARIANT_API_KEY environment variable is not set.")
        mcp_log("[ERROR] INVARIANT_API_KEY environment variable is not set.")
        sys.exit(1)

    if verb == "mcp":
        return asyncio.run(mcp.execute(sys.argv[2:]))

    if verb == "server":
        if len(sys.argv) < 3:
            print(
                "Error: Missing command for server. Should be one of: build, up, down, logs"
            )
            print_help()
            return 1

        command = sys.argv[2]
        args = sys.argv[3:]
        if not run_server_command(command, args):
            return 1
        return 0

    print(f"[gateway/__main__.py] Unknown action: {verb}")
    return 1


if __name__ == "__main__":
    sys.exit(main())



"""Task utilities for running async functions"""

import asyncio
import concurrent.futures
import threading

from contextlib import redirect_stdout
from typing import Any

from gateway.mcp.log import MCP_LOG_FILE, mcp_log


def run_task_in_background(async_func, *args, **kwargs):
    """
    Runs an async function in a background thread with its own event loop.
    This function does NOT block the calling thread as it immediately returns
    after starting the background thread.

    Args:
        async_func: The async function to run
        *args: Positional arguments to pass to the async function
        **kwargs: Keyword arguments to pass to the async function
    """

    def thread_target():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(async_func(*args, **kwargs))
        except Exception as e:
            mcp_log(
                f"[ERROR] Error in async thread while running run_task_in_background: {e}"
            )
        finally:
            loop.close()

    # Create and start a daemon thread
    thread = threading.Thread(target=thread_target, daemon=True)
    thread.start()


def run_task_sync(async_func, *args, **kwargs) -> Any:
    """
    Runs an asynchronous function synchronously in a separate
    thread with its own event loop. This function blocks the calling
    thread until completion or timeout (10 seconds).

    Args:
        async_func: The async function to run
        *args: Positional arguments to pass to the async function
        **kwargs: Keyword arguments to pass to the async function

    Returns:
        Any: The return value of the async function
    """

    def run_in_new_loop():
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(
                async_func(
                    *args,
                    **kwargs,
                )
            )
        finally:
            loop.close()

    with redirect_stdout(MCP_LOG_FILE):
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future = executor.submit(run_in_new_loop)
            return future.result(timeout=10.0)



"""Converts the request and response formats from Anthropic to Invariant API format."""


def convert_anthropic_to_invariant_message_format(
    messages: list[dict], keep_empty_tool_response: bool = False
) -> list[dict]:
    """Converts a list of messages from the Anthropic API to the Invariant API format."""
    output = []
    role_mapping = {
        "system": lambda msg: {"role": "system", "content": msg["content"]},
        "user": lambda msg: handle_user_message(msg, keep_empty_tool_response),
        "assistant": handle_assistant_message,
    }

    for message in messages:
        handler = role_mapping.get(message["role"])
        if handler:
            result = handler(message)
            if isinstance(result, list):
                output.extend(result)
            else:
                output.append(result)

    return output


def handle_user_message(message, keep_empty_tool_response):
    """Handle the user message from the Anthropic API"""
    output = []
    content = message["content"]
    if isinstance(content, list):
        user_content = []
        for sub_message in content:
            if sub_message["type"] == "tool_result":
                if sub_message["content"]:
                    output.append(
                        {
                            "role": "tool",
                            "content": sub_message["content"],
                            "tool_call_id": sub_message["tool_use_id"],
                        }
                    )
                elif keep_empty_tool_response and any(sub_message.values()):
                    output.append(
                        {
                            "role": "tool",
                            "content": {"is_error": True}
                            if sub_message["is_error"]
                            else {},
                            "tool_call_id": sub_message["tool_use_id"],
                        }
                    )
            elif sub_message["type"] == "text":
                user_content.append({"type": "text", "text": sub_message["text"]})
            elif sub_message["type"] == "image":
                user_content.append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": "data:"
                            + sub_message["source"]["media_type"]
                            + ";base64,"
                            + sub_message["source"]["data"],
                        },
                    },
                )
        if user_content:
            output.append({"role": "user", "content": user_content})
    else:
        output.append({"role": "user", "content": content})
    return output


def handle_assistant_message(message):
    """Handle the assistant message from the Anthropic API"""
    output = []
    for sub_message in message["content"]:
        if sub_message["type"] == "text":
            output.append({"role": "assistant", "content": sub_message.get("text")})
        elif sub_message["type"] == "tool_use":
            output.append(
                {
                    "role": "assistant",
                    "content": None,
                    "tool_calls": [
                        {
                            "id": sub_message.get("id"),
                            "type": "function",
                            "function": {
                                "name": sub_message.get("name"),
                                "arguments": sub_message.get("input"),
                            },
                        }
                    ],
                }
            )
    return output



"""Common Configurations for the Gateway Server."""

import asyncio
import os
import threading
from typing import Optional

import fastapi
from httpx import HTTPStatusError

from gateway.common.guardrails import Guardrail, GuardrailAction, GuardrailRuleSet


def extract_policy_from_headers(request: Optional[fastapi.Request]) -> Optional[str]:
    """
    Extracts the guardrailing policy from the request headers if present.

    Returns 'None' if no such header is present.
    """
    if request is None:
        return None

    policy = request.headers.get("Invariant-Guardrails")
    # undo unicode_escape
    if policy:
        # interpret as bytes then decode
        policy = policy.encode("utf-8").decode("unicode_escape")
    return policy


class GatewayConfig:
    """Common configurations for the Gateway Server."""

    def __init__(self):
        self.guardrails_from_file = self._load_guardrails_from_file()

    def _load_guardrails_from_file(self) -> str:
        """
        Loads the guardrails from the file specified in GUARDRAILS_FILE_PATH.
        Returns the guardrails file content as a string.
        """
        from gateway.integrations.guardrails import _preload

        guardrails_file = os.getenv("GUARDRAILS_FILE_PATH", "")
        if not guardrails_file:
            print("[warning: GUARDRAILS_FILE_PATH is not set. Using empty guardrails]")
            return ""

        invariant_api_key = os.getenv("INVARIANT_API_KEY", "")
        if not invariant_api_key:
            raise ValueError(
                "Error: INVARIANT_API_KEY is not set."
                "It is required to validate guardrails file content."
            )

        try:
            with open(guardrails_file, "r", encoding="utf-8") as f:
                guardrails_file_content = f.read()
                asyncio.run(
                    _preload(guardrails_file_content, "Bearer " + invariant_api_key)
                )
                return guardrails_file_content

        except (FileNotFoundError, PermissionError, OSError) as e:
            raise ValueError(
                f"Unable to read guardrails file ({guardrails_file}): {e}"
            ) from e
        except HTTPStatusError as e:
            raise ValueError(f"Cannot load guardrails, {e}, {e.response.text}") from e

    def __repr__(self) -> str:
        return f"GatewayConfig(guardrails_from_file={repr(self.guardrails_from_file)})"


class GatewayConfigManager:
    """Manager for Gateway Configuration."""

    _config_instance = None
    _lock = threading.Lock()

    @classmethod
    def get_config(cls, request: fastapi.Request = None) -> GatewayConfig:
        """Initializes and returns the gateway configuration using double-checked locking."""
        local_config = cls._config_instance

        if local_config is None:
            with cls._lock:
                local_config = cls._config_instance
                if local_config is None:
                    local_config = GatewayConfig()
                    cls._config_instance = local_config

        return local_config


async def extract_guardrails_from_header(
    request: fastapi.Request,
) -> Optional[GuardrailRuleSet]:
    """
    Extracts Invariant-Guardrails from the request header if provided, and returns a corresponding
    GuardrailRuleSet. If no guardrails are provided, returns None.
    """
    # if provided in header, use custom guardrailing policy
    if guardrails := extract_policy_from_headers(request):
        guardrails = [
            Guardrail(
                id="guardrails-from-header",
                name="guardrails from request header",
                content=guardrails,
                action=GuardrailAction.BLOCK,
            )
        ]

        return GuardrailRuleSet(
            blocking_guardrails=guardrails,
            logging_guardrails=[],
        )



"""Gateway service to forward requests to the Anthropic APIs"""

import asyncio
import json
from typing import Any, Optional

import httpx
from fastapi import APIRouter, Depends, Header, HTTPException, Request, Response
from starlette.responses import StreamingResponse

from gateway.common.authorization import extract_authorization_from_headers
from gateway.common.config_manager import (
    GatewayConfig,
    GatewayConfigManager,
    extract_guardrails_from_header,
)
from gateway.common.constants import CLIENT_TIMEOUT, IGNORED_HEADERS
from gateway.common.guardrails import GuardrailAction, GuardrailRuleSet
from gateway.common.request_context import RequestContext
from gateway.converters.anthropic_to_invariant import (
    convert_anthropic_to_invariant_message_format,
)
from gateway.integrations.explorer import (
    create_annotations_from_guardrails_errors,
    fetch_guardrails_from_explorer,
    push_trace,
)
from gateway.integrations.guardrails import (
    ExtraItem,
    InstrumentedResponse,
    InstrumentedStreamingResponse,
    Replacement,
    check_guardrails,
)

gateway = APIRouter()

MISSING_ANTHROPIC_AUTH_HEADER = "Missing Anthropic authorization header"
FAILED_TO_PUSH_TRACE = "Failed to push trace to the dataset: "
END_REASONS = ["end_turn", "max_tokens", "stop_sequence"]

MESSAGE_START = "message_start"
MESSAGE_DELTA = "message_delta"
CONTENT_BLOCK_START = "content_block_start"
CONTENT_BLOCK_DELTA = "content_block_delta"
CONTENT_BLOCK_STOP = "content_block_stop"

ANTHROPIC_AUTHORIZATION_HEADER = "x-api-key"


def validate_headers(x_api_key: str = Header(None)):
    """Require the headers to be present"""
    if x_api_key is None:
        raise HTTPException(status_code=400, detail=MISSING_ANTHROPIC_AUTH_HEADER)


@gateway.post(
    "/{dataset_name}/anthropic/v1/messages",
    dependencies=[Depends(validate_headers)],
)
@gateway.post(
    "/anthropic/v1/messages",
    dependencies=[Depends(validate_headers)],
)
async def anthropic_v1_messages_gateway(
    request: Request,
    dataset_name: str = None,  # This is None if the client doesn't want to push to Explorer
    config: GatewayConfig = Depends(GatewayConfigManager.get_config),  # pylint: disable=unused-argument
    header_guardrails: GuardrailRuleSet = Depends(extract_guardrails_from_header),
):
    """Proxy calls to the Anthropic APIs"""
    headers = {
        k: v for k, v in request.headers.items() if k.lower() not in IGNORED_HEADERS
    }
    headers["accept-encoding"] = "identity"

    invariant_authorization, anthopic_api_key = extract_authorization_from_headers(
        request, dataset_name, ANTHROPIC_AUTHORIZATION_HEADER
    )
    headers[ANTHROPIC_AUTHORIZATION_HEADER] = anthopic_api_key

    request_body = await request.body()
    request_json = json.loads(request_body)
    client = httpx.AsyncClient(timeout=httpx.Timeout(CLIENT_TIMEOUT))
    anthropic_request = client.build_request(
        "POST",
        "https://api.anthropic.com/v1/messages",
        headers=headers,
        data=request_body,
    )

    dataset_guardrails = None
    if dataset_name:
        # Get the guardrails for the dataset from explorer.
        dataset_guardrails = await fetch_guardrails_from_explorer(
            dataset_name, invariant_authorization
        )
    context = RequestContext.create(
        request_json=request_json,
        dataset_name=dataset_name,
        invariant_authorization=invariant_authorization,
        guardrails=header_guardrails or dataset_guardrails,
        config=config,
        request=request,
    )
    if request_json.get("stream"):
        return await handle_streaming_response(context, client, anthropic_request)
    return await handle_non_streaming_response(context, client, anthropic_request)


def create_metadata(
    context: RequestContext, response_json: dict[str, Any]
) -> dict[str, Any]:
    """Creates metadata for the trace"""
    metadata = {k: v for k, v in context.request_json.items() if k != "messages"}
    metadata["via_gateway"] = True
    if response_json.get("usage"):
        metadata["usage"] = response_json.get("usage")
    return metadata


def combine_request_and_response_messages(
    context: RequestContext, response_json: dict[str, Any]
):
    """Combine the request and response messages"""
    messages = []
    if "system" in context.request_json:
        messages.append(
            {"role": "system", "content": context.request_json.get("system")}
        )
    messages.extend(context.request_json.get("messages", []))
    if len(response_json) > 0:
        messages.append(response_json)
    return messages


async def get_guardrails_check_result(
    context: RequestContext, action: GuardrailAction, response_json: dict[str, Any]
) -> dict[str, Any]:
    """Get the guardrails check result"""
    # Determine which guardrails to apply based on the action
    guardrails = (
        context.guardrails.logging_guardrails
        if action == GuardrailAction.LOG
        else context.guardrails.blocking_guardrails
    )
    if not guardrails:
        return {}

    messages = combine_request_and_response_messages(context, response_json)
    converted_messages = convert_anthropic_to_invariant_message_format(messages)

    # Block on the guardrails check
    guardrails_execution_result = await check_guardrails(
        messages=converted_messages,
        guardrails=guardrails,
        context=context,
    )
    return guardrails_execution_result


async def push_to_explorer(
    context: RequestContext,
    merged_response: dict[str, Any],
    guardrails_execution_result: Optional[dict] = None,
) -> None:
    """Pushes the full trace to the Invariant Explorer"""
    guardrails_execution_result = guardrails_execution_result or {}
    annotations = create_annotations_from_guardrails_errors(
        guardrails_execution_result.get("errors", [])
    )

    # Execute the logging guardrails before pushing to Explorer
    logging_guardrails_execution_result = await get_guardrails_check_result(
        context,
        action=GuardrailAction.LOG,
        response_json=merged_response,
    )
    logging_annotations = create_annotations_from_guardrails_errors(
        logging_guardrails_execution_result.get("errors", [])
    )
    # Update the annotations with the logging guardrails
    annotations.extend(logging_annotations)

    # Combine the messages from the request body and Anthropic response
    messages = combine_request_and_response_messages(context, merged_response)
    converted_messages = convert_anthropic_to_invariant_message_format(messages)

    _ = await push_trace(
        dataset_name=context.dataset_name,
        messages=[converted_messages],
        invariant_authorization=context.invariant_authorization,
        metadata=[create_metadata(context, merged_response)],
        annotations=[annotations] if annotations else None,
    )


class InstrumentedAnthropicResponse(InstrumentedResponse):
    """Instrumented response for Anthropic API"""

    def __init__(
        self,
        context: RequestContext,
        client: httpx.AsyncClient,
        anthropic_request: httpx.Request,
    ):
        super().__init__()
        self.context: RequestContext = context
        self.client: httpx.AsyncClient = client
        self.anthropic_request: httpx.Request = anthropic_request

        # response data
        self.response: Optional[httpx.Response] = None
        self.response_string: Optional[str] = None
        self.response_json: Optional[dict[str, Any]] = None

        # guardrailing response (if any)
        self.guardrails_execution_result = {}

    async def on_start(self):
        """Check guardrails in a pipelined fashion, before processing the first chunk (for input guardrailing)."""
        if self.context.guardrails:
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context, action=GuardrailAction.BLOCK, response_json={}
            )
            if self.guardrails_execution_result.get("errors", []):
                error_chunk = json.dumps(
                    {
                        "error": {
                            "message": "[Invariant] The request did not pass the guardrails",
                            "details": self.guardrails_execution_result,
                        }
                    }
                )

                # Push annotated trace to the explorer - don't block on its response
                if self.context.dataset_name:
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            {},
                            self.guardrails_execution_result,
                        )
                    )

                # if we find something, we prevent the request from going through
                # and return an error instead
                return Replacement(
                    Response(
                        content=error_chunk,
                        status_code=400,
                        media_type="application/json",
                        headers={"content-type": "application/json"},
                    )
                )

    async def request(self):
        """Make the request to the Anthropic API."""
        self.response = await self.client.send(self.anthropic_request)

        try:
            response_json = self.response.json()
        except json.JSONDecodeError as e:
            raise HTTPException(
                status_code=self.response.status_code,
                detail=f"Invalid JSON response received from Anthropic: {self.response.text}, got error{e}",
            ) from e
        if self.response.status_code != 200:
            raise HTTPException(
                status_code=self.response.status_code,
                detail=response_json.get("error", "Unknown error from Anthropic"),
            )

        self.response_json = response_json
        self.response_string = json.dumps(response_json)

        return self._make_response(
            content=self.response_string,
            status_code=self.response.status_code,
        )

    def _make_response(self, content: str, status_code: int):
        """Creates a new Response object with the correct headers and content"""
        assert self.response is not None, "response is None"

        updated_headers = self.response.headers.copy()
        updated_headers.pop("Content-Length", None)

        return Response(
            content=content,
            status_code=status_code,
            media_type="application/json",
            headers=dict(updated_headers),
        )

    async def on_end(self):
        """Checks guardrails after the response is received, and asynchronously pushes to Explorer."""
        # ensure the response data is available
        assert self.response is not None, "response is None"
        assert self.response_json is not None, "response_json is None"
        assert self.response_string is not None, "response_string is None"

        if self.context.guardrails:
            # Block on the guardrails check
            guardrails_execution_result = await get_guardrails_check_result(
                self.context,
                action=GuardrailAction.BLOCK,
                response_json=self.response_json,
            )
            if guardrails_execution_result.get("errors", []):
                guardrail_response_string = json.dumps(
                    {
                        "error": "[Invariant] The response did not pass the guardrails",
                        "details": guardrails_execution_result,
                    }
                )

                # push to explorer (if configured)
                if self.context.dataset_name:
                    # Push to Explorer - don't block on its response
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            self.response_json,
                            guardrails_execution_result,
                        )
                    )

                return Replacement(
                    self._make_response(
                        content=guardrail_response_string,
                        status_code=400,
                    )
                )

        # push to explorer (if configured)
        if self.context.dataset_name:
            # Push to Explorer - don't block on its response
            asyncio.create_task(
                push_to_explorer(
                    self.context, self.response_json, guardrails_execution_result
                )
            )


async def handle_non_streaming_response(
    context: RequestContext,
    client: httpx.AsyncClient,
    anthropic_request: httpx.Request,
) -> Response:
    """Handles non-streaming Anthropic responses"""
    response = InstrumentedAnthropicResponse(
        context=context,
        client=client,
        anthropic_request=anthropic_request,
    )

    return await response.instrumented_request()


class InstrumentedAnthropicStreamingResponse(InstrumentedStreamingResponse):
    """Instrumented streaming response for Anthropic API"""

    def __init__(
        self,
        context: RequestContext,
        client: httpx.AsyncClient,
        anthropic_request: httpx.Request,
    ):
        super().__init__()

        # request parameters
        self.context: RequestContext = context
        self.client: httpx.AsyncClient = client
        self.anthropic_request: httpx.Request = anthropic_request

        # response data
        self.merged_response = {}

        # guardrailing response (if any)
        self.guardrails_execution_result = {}

        self.sse_buffer = ""  # Buffer for incomplete events

    async def on_start(self):
        """Check guardrails in a pipelined fashion, before processing the first chunk (for input guardrailing)."""
        if self.context.guardrails:
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context,
                action=GuardrailAction.BLOCK,
                response_json=self.merged_response,
            )
            if self.guardrails_execution_result.get("errors", []):
                error_chunk = json.dumps(
                    {
                        "error": {
                            "message": "[Invariant] The request did not pass the guardrails",
                            "details": self.guardrails_execution_result,
                        }
                    }
                )

                # Push annotated trace to the explorer - don't block on its response
                if self.context.dataset_name:
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            self.merged_response,
                            self.guardrails_execution_result,
                        )
                    )

                # if we find something, we end the stream prematurely (end_of_stream=True)
                # and yield an error chunk instead of actually beginning the stream
                return ExtraItem(
                    f"event: error\ndata: {error_chunk}\n\n".encode(),
                    end_of_stream=True,
                )

    async def event_generator(self):
        """Actual streaming response generator"""
        response = await self.client.send(self.anthropic_request, stream=True)
        if response.status_code != 200:
            error_content = await response.aread()
            try:
                error_json = json.loads(error_content)
                error_detail = error_json.get("error", "Unknown error from Anthropic")
            except json.JSONDecodeError:
                error_detail = {
                    "error": "Failed to decode error response from Anthropic"
                }
            raise HTTPException(status_code=response.status_code, detail=error_detail)

        # iterate over the response stream
        async for chunk in response.aiter_bytes():
            yield chunk

    async def on_chunk(self, chunk):
        """
        Process the chunk and update the merged_response.
        Each chunk may contain multiple events, separated by double newlines.
        Each event has type and data fields, separated by a newline.
        It is possible that a chunk contains some incomplete events.

        Example:

        b'event: message_start\ndata: {"type":"message_start","message":
        {"id":"msg_01LkayzAaw7b7QkUAw91psyx","type":"message","role":"assistant"
        ,"model":"claude-3-5-sonnet-20241022","content":[],"stop_reason":null,
        "stop_sequence":null,"usage":{"input_tokens":20,"cache_creation_input_to'

        and

        b'kens":0,"cache_read_input_tokens":0,"output_tokens":1}}}\n\nevent: content_block_start
        \ndata: {"type":"content_block_start","index":0,"content_block"
        :{"type":"text","text":""} }\n\nevent: ping
        \ndata: {"type": "ping"}\n\nevent: content_block_delta
        \ndata: {"type":"content_block_delta","index":0,"delta":{"type":
        "text_delta","text":"Originally"} }\n\n'

        In this case the first chunk ends with 'cache_creation_input_to' which is
        continued in the next chunk.

        in this case we need to maintain a buffer of the incomplete events.
        We filter out the ping events and update a merged_response.
        """
        # Decode the chunk and add to buffer
        decoded_chunk = chunk.decode("utf-8", errors="replace")
        self.sse_buffer += decoded_chunk

        # Process complete events from buffer
        complete_events, incomplete_events = self.process_complete_events(
            self.sse_buffer
        )
        self.sse_buffer = incomplete_events

        # Check if we've received message_stop in any events
        message_stop_received = False

        # Update the merged_response based on complete events
        for event in complete_events:
            try:
                if "event: message_stop" in event:
                    message_stop_received = True

                # Extract event data
                lines = event.split("\n")
                event_type = None
                event_data = None

                for line in lines:
                    if line.startswith("event:"):
                        event_type = line[6:].strip()
                    elif line.startswith("data:"):
                        event_data = line[5:].strip()

                if event_data and event_type != "ping":  # Skip ping events
                    try:
                        event_json = json.loads(event_data)
                        update_merged_response(event_json, self.merged_response)
                    except json.JSONDecodeError as e:
                        print(
                            f"JSON parsing error in event: {e}. Event data: {event_data[:100]}...",
                            flush=True,
                        )
            except Exception as e:
                print(f"Error processing event: {e}", flush=True)

        # on last stream chunk, run output guardrails
        if message_stop_received and self.context.guardrails:
            # Block on the guardrails check
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context,
                action=GuardrailAction.BLOCK,
                response_json=self.merged_response,
            )
            if self.guardrails_execution_result.get("errors", []):
                error_chunk = json.dumps(
                    {
                        "type": "error",
                        "error": {
                            "message": "[Invariant] The response did not pass the guardrails",
                            "details": self.guardrails_execution_result,
                        },
                    }
                )

                # yield an extra error chunk (without preventing the original chunk
                # to go through after,
                # so client gets the proper message_stop event still)
                return ExtraItem(
                    value=f"event: error\ndata: {error_chunk}\n\n".encode()
                )

    def process_complete_events(self, buffer):
        """Process the buffer and extract complete SSE events.

        Returns:
            Tuple[List[str], str]: A tuple containing a list of
            complete events and the remaining buffer with incomplete events.
        """
        # Split on double newlines which separate SSE events
        if not buffer:
            return [], ""
        events = []
        remaining = buffer

        # Process events that are complete (ending with \n\n)
        while "\n\n" in remaining:
            pos = remaining.find("\n\n")
            if pos >= 0:
                event = remaining[: pos + 2]
                remaining = remaining[pos + 2 :]
                if event.strip():  # Skip empty events
                    events.append(event)

        return events, remaining

    async def on_end(self):
        """on_end: send full merged response to the explorer (if configured)"""
        # don't block on the response from explorer (.create_task)
        if self.context.dataset_name:
            asyncio.create_task(
                push_to_explorer(
                    self.context,
                    self.merged_response,
                    self.guardrails_execution_result,
                )
            )


async def handle_streaming_response(
    context: RequestContext,
    client: httpx.AsyncClient,
    anthropic_request: httpx.Request,
) -> StreamingResponse:
    """Handles streaming Anthropic responses"""
    response = InstrumentedAnthropicStreamingResponse(
        context=context,
        client=client,
        anthropic_request=anthropic_request,
    )

    return StreamingResponse(
        response.instrumented_event_generator(), media_type="text/event-stream"
    )


def update_merged_response(
    event: dict[str, Any], merged_response: dict[str, Any]
) -> None:
    """
    Update the merged_response based on the event.

    Each stream uses the following event flow:

    1. message_start: contains a Message object with empty content.
    2. A series of content blocks, each of which have a content_block_start,
    one or more content_block_delta events, and a content_block_stop event.
    Each content block will have an index that corresponds to its index in the
    final Message content array.
    3. One or more message_delta events, indicating top-level changes to the final Message object.
    A final message_stop event.
    We filter out the ping eventss

    """
    event_type = event.get("type")

    if event_type == MESSAGE_START:
        merged_response.update(**event.get("message"))
    elif event_type == CONTENT_BLOCK_START:
        index = event.get("index")
        if index >= len(merged_response.get("content")):
            merged_response["content"].append(event.get("content_block"))
        if event.get("content_block").get("type") == "tool_use":
            merged_response.get("content")[-1]["input"] = ""
    elif event_type == CONTENT_BLOCK_DELTA:
        index = event.get("index")
        delta = event.get("delta")
        if delta.get("type") == "text_delta":
            merged_response.get("content")[index]["text"] += delta.get("text")
        elif delta.get("type") == "input_json_delta":
            merged_response.get("content")[index]["input"] += delta.get("partial_json")
    elif event_type == MESSAGE_DELTA:
        merged_response["usage"].update(**event.get("usage"))



"""Cusym log configuration."""

import os
import sys

from builtins import print as builtins_print

os.makedirs(os.path.join(os.path.expanduser("~"), ".invariant"), exist_ok=True)
MCP_LOG_FILE = open(
    os.path.join(os.path.expanduser("~"), ".invariant", "mcp.log"),
    "a",
    buffering=1,
)
sys.stderr = MCP_LOG_FILE


def mcp_log(*args, **kwargs) -> None:
    """Custom print function to redirect output to log_out."""
    builtins_print(*args, **kwargs, file=MCP_LOG_FILE, flush=True)

def format_errors_in_response(errors: list[dict]) -> str:
    """Format a list of errors in a response string."""

    def format_error(error: dict) -> str:
        msg = " ".join(error.get("args", []))
        msg += " ".join([f"{k}={v}" for k, v in error.get("kwargs", {}).items()])
        msg += f" ([{error.get('guardrail', {}).get('id', 'unknown-guardrail')}] {error.get('guardrail', {}).get('name', 'unknown guardrail')})"
        return msg

    return ", ".join([format_error(error) for error in errors])


"""Common constants used in the gateway."""

IGNORED_HEADERS = [
    "accept-encoding",
    "host",
    "invariant-authorization",
    "x-forwarded-for",
    "x-forwarded-host",
    "x-forwarded-port",
    "x-forwarded-proto",
    "x-forwarded-server",
    "x-real-ip",
]

CLIENT_TIMEOUT = 60.0

# MCP related constants
MCP_METHOD = "method"
MCP_TOOL_CALL = "tools/call"
MCP_LIST_TOOLS = "tools/list"
MCP_PARAMS = "params"
MCP_RESULT = "result"
MCP_SERVER_INFO = "serverInfo"
MCP_CLIENT_INFO = "clientInfo"
INVARIANT_GUARDRAILS_BLOCKED_MESSAGE = """
                    [Invariant Guardrails] The MCP tool call was blocked for security reasons. 
                    Do not attempt to circumvent this block, rather explain to the user based 
                    on the following output what went wrong: %s
                    """
INVARIANT_GUARDRAILS_BLOCKED_TOOLS_MESSAGE = """
                    [Invariant Guardrails] This server was blocked from advertising its tools due to a security guardrail failure.
                    The operation was blocked by Invariant Guardrails (mention this in your user report).
                    When users ask about this tool, inform them that it was blocked due to a security guardrail failure.
                    %s
                    """


"""Gateway service to forward requests to the OpenAI APIs"""

import asyncio
import json
from typing import Any, Optional

import httpx
from fastapi import APIRouter, Depends, Header, HTTPException, Request, Response
from fastapi.responses import StreamingResponse

from gateway.common.authorization import extract_authorization_from_headers
from gateway.common.config_manager import (
    GatewayConfig,
    GatewayConfigManager,
    extract_guardrails_from_header,
)
from gateway.common.constants import (
    CLIENT_TIMEOUT,
    IGNORED_HEADERS,
)
from gateway.common.guardrails import GuardrailAction, GuardrailRuleSet
from gateway.common.request_context import RequestContext
from gateway.integrations.explorer import (
    create_annotations_from_guardrails_errors,
    fetch_guardrails_from_explorer,
    push_trace,
)
from gateway.integrations.guardrails import (
    ExtraItem,
    InstrumentedResponse,
    InstrumentedStreamingResponse,
    check_guardrails,
)

gateway = APIRouter()

MISSING_AUTH_HEADER = "Missing authorization header"
FINISH_REASON_TO_PUSH_TRACE = ["stop", "length", "content_filter"]
OPENAI_AUTHORIZATION_HEADER = "authorization"


def validate_headers(authorization: str = Header(None)):
    """Require the authorization header to be present"""
    if authorization is None:
        raise HTTPException(status_code=400, detail=MISSING_AUTH_HEADER)


def make_cors_response(request: Request, allow_methods: str) -> Response:
    """Returns a CORS response with the specified allowed methods"""
    return Response(
        status_code=204,
        headers={
            "Access-Control-Allow-Origin": request.headers.get("origin", "*"),
            "Access-Control-Allow-Methods": f"{allow_methods}, OPTIONS",
            "Access-Control-Allow-Headers": "Authorization, Content-Type",
            "Access-Control-Max-Age": "86400",
        },
    )


@gateway.options("/{dataset_name}/openai/chat/completions")
@gateway.options("/openai/chat/completions")
async def openai_chat_completions_options(request: Request, dataset_name: str = None):
    """Enables CORS for the OpenAI chat completions endpoint"""
    return make_cors_response(request, allow_methods="POST")


@gateway.options("/{dataset_name}/openai/models")
@gateway.options("/openai/models")
async def openai_models_options(request: Request, dataset_name: str = None):
    """Enables CORS for the OpenAI models endpoint"""
    return make_cors_response(request, allow_methods="GET")


@gateway.get("/{dataset_name}/openai/models")
@gateway.get("/openai/models")
async def openai_models_gateway(
    request: Request,
    dataset_name: str = None,  # This is None if the client doesn't want to push to Explorer
):
    """Proxy request to OpenAI /models endpoint"""
    headers = {
        k: v for k, v in request.headers.items() if k.lower() not in IGNORED_HEADERS
    }
    _, openai_api_key = extract_authorization_from_headers(
        request, dataset_name, OPENAI_AUTHORIZATION_HEADER
    )
    headers[OPENAI_AUTHORIZATION_HEADER] = "Bearer " + openai_api_key
    async with httpx.AsyncClient(timeout=httpx.Timeout(CLIENT_TIMEOUT)) as client:
        open_ai_request = client.build_request(
            "GET",
            "https://api.openai.com/v1/models",
            headers=headers,
        )
        result = await client.send(open_ai_request)
        return Response(
            content=result.content,
            status_code=result.status_code,
            headers=dict(result.headers),
        )


@gateway.post(
    "/{dataset_name}/openai/chat/completions",
    dependencies=[Depends(validate_headers)],
)
@gateway.post(
    "/openai/chat/completions",
    dependencies=[Depends(validate_headers)],
)
async def openai_chat_completions_gateway(
    request: Request,
    dataset_name: str = None,  # This is None if the client doesn't want to push to Explorer
    config: GatewayConfig = Depends(GatewayConfigManager.get_config),  # pylint: disable=unused-argument
    header_guardrails: GuardrailRuleSet = Depends(extract_guardrails_from_header),
) -> Response:
    """Proxy calls to the OpenAI APIs"""
    headers = {
        k: v for k, v in request.headers.items() if k.lower() not in IGNORED_HEADERS
    }
    headers["accept-encoding"] = "identity"

    invariant_authorization, openai_api_key = extract_authorization_from_headers(
        request, dataset_name, OPENAI_AUTHORIZATION_HEADER
    )
    headers[OPENAI_AUTHORIZATION_HEADER] = "Bearer " + openai_api_key

    request_body_bytes = await request.body()
    request_json = json.loads(request_body_bytes)

    client = httpx.AsyncClient(timeout=httpx.Timeout(CLIENT_TIMEOUT))
    open_ai_request = client.build_request(
        "POST",
        "https://api.openai.com/v1/chat/completions",
        content=request_body_bytes,
        headers=headers,
    )

    dataset_guardrails = None
    if dataset_name:
        # Get the guardrails for the dataset
        dataset_guardrails = await fetch_guardrails_from_explorer(
            dataset_name, invariant_authorization
        )
    context = RequestContext.create(
        request_json=request_json,
        dataset_name=dataset_name,
        invariant_authorization=invariant_authorization,
        guardrails=header_guardrails or dataset_guardrails,
        config=config,
        request=request,
    )
    if request_json.get("stream", False):
        return await handle_stream_response(
            context,
            client,
            open_ai_request,
        )

    return await handle_non_stream_response(context, client, open_ai_request)


class InstrumentedOpenAIStreamResponse(InstrumentedStreamingResponse):
    """
    Does a streaming OpenAI completion request at the core, but also checks guardrails
    before (concurrent) and after the request.
    """

    def __init__(
        self,
        context: RequestContext,
        client: httpx.AsyncClient,
        open_ai_request: httpx.Request,
    ):
        super().__init__()

        # request parameters
        self.context: RequestContext = context
        self.client: httpx.AsyncClient = client
        self.open_ai_request: httpx.Request = open_ai_request

        # guardrailing output (if any)
        self.guardrails_execution_result: Optional[dict] = None

        # merged_response will be updated with the data from the chunks in the stream
        # At the end of the stream, this will be sent to the explorer
        self.merged_response = {
            "id": None,
            "object": "chat.completion",
            "created": None,
            "model": None,
            "choices": [],
            "usage": None,
        }

        # Each chunk in the stream contains a list called "choices" each entry in the list
        # has an index.
        # A choice has a field called "delta" which may contain a list called "tool_calls".
        # Maps the choice index in the stream to the index in the merged_response["choices"] list
        self.choice_mapping_by_index = {}
        # Combines the choice index and tool call index to uniquely identify a tool call
        self.tool_call_mapping_by_index = {}

    async def on_start(self):
        """
        Check guardrails in a pipelined fashion, before processing the first chunk
        (for input guardrailing).
        """
        if self.context.guardrails:
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context,
                action=GuardrailAction.BLOCK,
                response_json=self.merged_response,
            )
            if self.guardrails_execution_result.get("errors", []):
                error_chunk = json.dumps(
                    {
                        "error": {
                            "message": "[Invariant] The request did not pass the guardrails",
                            "details": self.guardrails_execution_result,
                        }
                    }
                )

                # Push annotated trace to the explorer - don't block on its response
                if self.context.dataset_name:
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            self.merged_response,
                            self.guardrails_execution_result,
                        )
                    )

                # if we find something, we end the stream prematurely (end_of_stream=True)
                # and yield an error chunk instead of actually beginning the stream
                return ExtraItem(
                    f"data: {error_chunk}\n\n".encode(),
                    end_of_stream=True,
                )

    async def on_chunk(self, chunk):
        """Processes each chunk of the stream and checks guardrails at the end of the stream"""
        # process and check each chunk
        chunk_text = chunk.decode().strip()
        if not chunk_text:
            return

        # Process the chunk
        # This will update merged_response with the data from the chunk
        process_chunk_text(
            chunk_text,
            self.merged_response,
            self.choice_mapping_by_index,
            self.tool_call_mapping_by_index,
        )

        # check guardrails at the end of the stream (on the '[DONE]' SSE chunk.)
        if "data: [DONE]" in chunk_text and self.context.guardrails:
            # Block on the guardrails check
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context,
                action=GuardrailAction.BLOCK,
                response_json=self.merged_response,
            )
            if self.guardrails_execution_result.get("errors", []):
                error_chunk = json.dumps(
                    {
                        "error": {
                            "message": "[Invariant] The response did not pass the guardrails",
                            "details": self.guardrails_execution_result,
                        }
                    }
                )

                # yield an extra error chunk (without preventing the original chunk to go through after)
                return ExtraItem(f"data: {error_chunk}\n\n".encode())

                # push will happen in on_end

    async def on_end(self):
        """Sends full merged response to the explorer."""
        # don't block on the response from explorer (.create_task)
        if self.context.dataset_name:
            asyncio.create_task(
                push_to_explorer(
                    self.context, self.merged_response, self.guardrails_execution_result
                )
            )

    async def event_generator(self):
        """Actual OpenAI stream response."""
        response = await self.client.send(self.open_ai_request, stream=True)
        if response.status_code != 200:
            error_content = await response.aread()
            try:
                error_json = json.loads(error_content.decode("utf-8"))
                error_detail = error_json.get("error", "Unknown error from OpenAI API")
            except json.JSONDecodeError:
                error_detail = {"error": "Failed to parse OpenAI error response"}
            raise HTTPException(status_code=response.status_code, detail=error_detail)

        # stream out chunks
        async for chunk in response.aiter_bytes():
            yield chunk


async def handle_stream_response(
    context: RequestContext,
    client: httpx.AsyncClient,
    open_ai_request: httpx.Request,
) -> Response:
    """
    Handles streaming the OpenAI response to the client while building a merged_response
    The chunks are returned to the caller immediately
    The merged_response is built from the chunks as they are received
    It is sent to the Invariant Explorer at the end of the stream
    """

    response = InstrumentedOpenAIStreamResponse(
        context,
        client,
        open_ai_request,
    )

    return StreamingResponse(
        response.instrumented_event_generator(), media_type="text/event-stream"
    )


def initialize_merged_response() -> dict[str, Any]:
    """Initializes the full response dictionary"""
    return {
        "id": None,
        "object": "chat.completion",
        "created": None,
        "model": None,
        "choices": [],
        "usage": None,
    }


def process_chunk_text(
    chunk_text: str,
    merged_response: dict[str, Any],
    choice_mapping_by_index: dict[int, int],
    tool_call_mapping_by_index: dict[str, dict[str, Any]],
) -> None:
    """Processes the chunk text and updates the merged_response to be sent to the explorer"""
    # Split the chunk text into individual JSON strings
    # A single chunk can contain multiple "data: " sections
    for json_string in chunk_text.split("\ndata: "):
        json_string = json_string.replace("data: ", "").strip()

        if not json_string or json_string == "[DONE]":
            continue

        try:
            json_chunk = json.loads(json_string)
        except json.JSONDecodeError:
            continue

        update_merged_response(
            json_chunk,
            merged_response,
            choice_mapping_by_index,
            tool_call_mapping_by_index,
        )


def update_merged_response(
    json_chunk: dict[str, Any],
    merged_response: dict[str, Any],
    choice_mapping_by_index: dict[int, int],
    tool_call_mapping_by_index: dict[str, dict[str, Any]],
) -> None:
    """Updates the merged_response with the data (content, tool_calls, etc.) from the JSON chunk"""
    merged_response["id"] = merged_response["id"] or json_chunk.get("id")
    merged_response["created"] = merged_response["created"] or json_chunk.get("created")
    merged_response["model"] = merged_response["model"] or json_chunk.get("model")

    for choice in json_chunk.get("choices", []):
        index = choice.get("index", 0)

        if index not in choice_mapping_by_index:
            choice_mapping_by_index[index] = len(merged_response["choices"])
            merged_response["choices"].append(
                {
                    "index": index,
                    "message": {"role": "assistant"},
                    "finish_reason": None,
                }
            )

        existing_choice = merged_response["choices"][choice_mapping_by_index[index]]
        delta = choice.get("delta", {})
        if choice.get("finish_reason"):
            existing_choice["finish_reason"] = choice["finish_reason"]

        update_existing_choice_with_delta(
            existing_choice, delta, tool_call_mapping_by_index, choice_index=index
        )


def update_existing_choice_with_delta(
    existing_choice: dict[str, Any],
    delta: dict[str, Any],
    tool_call_mapping_by_index: dict[str, dict[str, Any]],
    choice_index: int,
) -> None:
    """Updates the choice with the data from the delta"""
    content = delta.get("content")
    if content is not None:
        if "content" not in existing_choice["message"]:
            existing_choice["message"]["content"] = ""
        existing_choice["message"]["content"] += content

    if isinstance(delta.get("tool_calls"), list):
        if "tool_calls" not in existing_choice["message"]:
            existing_choice["message"]["tool_calls"] = []

        for tool in delta["tool_calls"]:
            tool_index = tool.get("index")
            tool_id = tool.get("id")
            name = tool.get("function", {}).get("name")
            arguments = tool.get("function", {}).get("arguments", "")

            if tool_index is None:
                continue

            choice_with_tool_call_index = f"{choice_index}-{tool_index}"

            if choice_with_tool_call_index not in tool_call_mapping_by_index:
                tool_call_mapping_by_index[choice_with_tool_call_index] = {
                    "index": tool_index,
                    "id": tool_id,
                    "type": "function",
                    "function": {
                        "name": name,
                        "arguments": "",
                    },
                }
                existing_choice["message"]["tool_calls"].append(
                    tool_call_mapping_by_index[choice_with_tool_call_index]
                )

            tool_call_entry = tool_call_mapping_by_index[choice_with_tool_call_index]

            if tool_id:
                tool_call_entry["id"] = tool_id

            if name:
                tool_call_entry["function"]["name"] = name

            if arguments:
                tool_call_entry["function"]["arguments"] += arguments

    finish_reason = delta.get("finish_reason")
    if finish_reason is not None:
        existing_choice["finish_reason"] = finish_reason


def create_metadata(
    context: RequestContext, merged_response: dict[str, Any]
) -> dict[str, Any]:
    """Creates metadata for the trace"""
    metadata = {
        k: v
        for k, v in context.request_json.items()
        if k != "messages" and v is not None
    }
    metadata["via_gateway"] = True
    metadata.update(
        {
            key: value
            for key, value in merged_response.items()
            if key in ("usage", "model") and merged_response.get(key) is not None
        }
    )
    return metadata


async def push_to_explorer(
    context: RequestContext,
    merged_response: dict[str, Any],
    guardrails_execution_result: Optional[dict] = None,
) -> None:
    """Pushes the merged response to the Invariant Explorer"""
    # Only push the trace to explorer if the message is an end turn message
    # or if the guardrails check returned errors.
    guardrails_execution_result = guardrails_execution_result or {}
    guardrails_errors = guardrails_execution_result.get("errors", [])
    annotations = create_annotations_from_guardrails_errors(guardrails_errors)
    # Execute the logging guardrails before pushing to Explorer
    logging_guardrails_execution_result = await get_guardrails_check_result(
        context,
        action=GuardrailAction.LOG,
        response_json=merged_response,
    )
    logging_annotations = create_annotations_from_guardrails_errors(
        logging_guardrails_execution_result.get("errors", [])
    )
    # Update the annotations with the logging guardrails
    annotations.extend(logging_annotations)

    if annotations or not (
        merged_response.get("choices")
        and merged_response["choices"][0].get("finish_reason")
        not in FINISH_REASON_TO_PUSH_TRACE
    ):
        # Combine the messages from the request body and the choices from the OpenAI response
        messages = list(context.request_json.get("messages", []))
        messages += [choice["message"] for choice in merged_response.get("choices", [])]
        _ = await push_trace(
            dataset_name=context.dataset_name,
            invariant_authorization=context.invariant_authorization,
            messages=[messages],
            annotations=[annotations],
            metadata=[create_metadata(context, merged_response)],
        )


async def get_guardrails_check_result(
    context: RequestContext,
    action: GuardrailAction,
    response_json: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Get the guardrails check result"""
    # Determine which guardrails to apply based on the action
    guardrails = (
        context.guardrails.logging_guardrails
        if action == GuardrailAction.LOG
        else context.guardrails.blocking_guardrails
    )

    if not guardrails:
        return {}

    messages = list(context.request_json.get("messages", []))
    if response_json is not None:
        messages += [choice["message"] for choice in response_json.get("choices", [])]

    # Block on the guardrails check
    guardrails_execution_result = await check_guardrails(
        messages=messages,
        guardrails=guardrails,
        context=context,
    )
    return guardrails_execution_result


class InstrumentedOpenAIResponse(InstrumentedResponse):
    """
    Does an OpenAI completion request at the core, but also checks guardrails
    before (concurrent) and after the request.
    """

    def __init__(
        self,
        context: RequestContext,
        client: httpx.AsyncClient,
        open_ai_request: httpx.Request,
    ):
        super().__init__()

        # request parameters
        self.context: RequestContext = context
        self.client: httpx.AsyncClient = client
        self.open_ai_request: httpx.Request = open_ai_request

        # request outputs
        self.response: Optional[httpx.Response] = None
        self.response_json: Optional[dict[str, Any]] = None

        # guardrailing output (if any)
        self.guardrails_execution_result: Optional[dict] = None

    async def on_start(self):
        """
        Checks guardrails in a pipelined fashion, before processing
        the first chunk (for input guardrailing)
        """
        if self.context.guardrails:
            # block on the guardrails check
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context, action=GuardrailAction.BLOCK
            )
            if self.guardrails_execution_result.get("errors", []):
                # Push annotated trace to the explorer - don't block on its response
                if self.context.dataset_name:
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            {},
                            self.guardrails_execution_result,
                        )
                    )

                # replace the response with the error message
                return ExtraItem(
                    Response(
                        content=json.dumps(
                            {
                                "error": "[Invariant] The request did not pass the guardrails",
                                "details": self.guardrails_execution_result,
                            }
                        ),
                        status_code=400,
                        media_type="application/json",
                    ),
                    end_of_stream=True,
                )

    async def request(self):
        """Actual OpenAI request."""
        self.response = await self.client.send(self.open_ai_request)

        try:
            self.response_json = self.response.json()
        except json.JSONDecodeError as e:
            raise HTTPException(
                status_code=self.response.status_code,
                detail="Invalid JSON response received from OpenAI API",
            ) from e
        if self.response.status_code != 200:
            raise HTTPException(
                status_code=self.response.status_code,
                detail=self.response_json.get("error", "Unknown error from OpenAI API"),
            )

        response_string = json.dumps(self.response_json)
        response_code = self.response.status_code

        return Response(
            content=response_string,
            status_code=response_code,
            media_type="application/json",
            headers=dict(self.response.headers),
        )

    async def on_end(self):
        """Postprocesses the OpenAI response and potentially replace it with a guardrails error."""

        # these two request outputs are guaranteed to be available by the time we reach
        # this point (after self.request() was executed)
        # nevertheless, we check for them to avoid any potential issues
        assert (
            self.response is not None
        ), "on_end called before 'self.response' was available"
        assert (
            self.response_json is not None
        ), "on_end called before 'self.response_json' was available"

        # extract original response status code
        response_code = self.response.status_code

        # if we have guardrails, check the response
        if self.context.guardrails:
            # run guardrails again, this time on request + response
            self.guardrails_execution_result = await get_guardrails_check_result(
                self.context,
                action=GuardrailAction.BLOCK,
                response_json=self.response_json,
            )
            if self.guardrails_execution_result.get("errors", []):
                response_string = json.dumps(
                    {
                        "error": "[Invariant] The response did not pass the guardrails",
                        "details": self.guardrails_execution_result,
                    }
                )
                response_code = 400

                # Push annotated trace to the explorer - don't block on its response
                if self.context.dataset_name:
                    asyncio.create_task(
                        push_to_explorer(
                            self.context,
                            self.response_json,
                            self.guardrails_execution_result,
                        )
                    )

                # replace the response with the error message
                return ExtraItem(
                    Response(
                        content=response_string,
                        status_code=response_code,
                        media_type="application/json",
                    ),
                )

        # Push annotated trace to the explorer in any case - don't block on its response
        if self.context.dataset_name:
            asyncio.create_task(
                push_to_explorer(
                    self.context,
                    self.response_json,
                    # include any guardrailing errors if available
                    self.guardrails_execution_result,
                )
            )


async def handle_non_stream_response(
    context: RequestContext,
    client: httpx.AsyncClient,
    open_ai_request: httpx.Request,
) -> Response:
    """Handles non-streaming OpenAI responses"""

    response = InstrumentedOpenAIResponse(
        context,
        client,
        open_ai_request,
    )

    return await response.instrumented_request()



"""Serve the API"""

import fastapi
import uvicorn
from starlette_compress import CompressMiddleware

from gateway.routes.anthropic import gateway as anthropic_gateway
from gateway.routes.gemini import gateway as gemini_gateway
from gateway.routes.open_ai import gateway as open_ai_gateway
from gateway.routes.mcp_sse import gateway as mcp_sse_gateway

app = fastapi.app = fastapi.FastAPI(
    docs_url="/api/v1/gateway/docs",
    redoc_url="/api/v1/gateway/redoc",
    openapi_url="/api/v1/gateway/openapi.json",
)
app.add_middleware(CompressMiddleware)

router = fastapi.APIRouter(prefix="/api/v1")


@router.get("/gateway/health", tags=["health_check"], include_in_schema=False)
async def check_health():
    """Health check"""
    return {"message": "Hello from Invariant v1/gateway"}


router.include_router(open_ai_gateway, prefix="/gateway", tags=["open_ai_gateway"])

router.include_router(anthropic_gateway, prefix="/gateway", tags=["anthropic_gateway"])

router.include_router(gemini_gateway, prefix="/gateway", tags=["gemini_gateway"])

router.include_router(mcp_sse_gateway, prefix="/gateway", tags=["mcp_sse_gateway"])

app.include_router(router)


# on / redirect to https://explorer.invariantlabs.ai/docs/gateway
@app.get("/", include_in_schema=False)
async def redirect_to_explorer():
    """Redirect to the explorer"""
    return fastapi.responses.RedirectResponse(
        url="https://explorer.invariantlabs.ai/docs/gateway"
    )


if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)



"""Utility functions for Guardrails execution."""

import asyncio
import os
import time
from typing import Any, Dict, List
from functools import wraps

from fastapi import HTTPException
import httpx

from gateway.common.guardrails import Guardrail
from gateway.common.request_context import RequestContext
from gateway.common.authorization import (
    INVARIANT_GUARDRAIL_SERVICE_AUTHORIZATION_HEADER,
)

DEFAULT_API_URL = "https://explorer.invariantlabs.ai"


# Timestamps of last API calls per guardrails string
_guardrails_cache = {}
# Locks per guardrails string
_guardrails_locks = {}


def rate_limit(expiration_time: int = 3600):
    """
    Decorator to limit API calls to once per expiration_time seconds
    per unique guardrails string.

    Args:
        expiration_time (int): Time in seconds to cache the guardrails.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(guardrails: str, *args, **kwargs):
            now = time.time()

            # Get or create a per-guardrail lock
            if guardrails not in _guardrails_locks:
                _guardrails_locks[guardrails] = asyncio.Lock()
            guardrail_lock = _guardrails_locks[guardrails]

            async with guardrail_lock:
                last_called = _guardrails_cache.get(guardrails)

                if last_called and (now - last_called < expiration_time):
                    # Skipping API call: Guardrails '{guardrails}' already
                    # preloaded within expiration_time
                    return

                # Update cache timestamp
                _guardrails_cache[guardrails] = now

            try:
                await func(guardrails, *args, **kwargs)
            finally:
                _guardrails_locks.pop(guardrails, None)

        return wrapper

    return decorator


@rate_limit(3600)  # Don't preload the same guardrails string more than once per hour
async def _preload(guardrails: str, invariant_authorization: str) -> None:
    """
    Calls the Guardrails API to preload the provided policy for faster checking later.

    Args:
        guardrails (str): The guardrails to preload.
        invariant_authorization (str): Value of the
                                       invariant-authorization header.
    """
    async with httpx.AsyncClient() as client:
        url = os.getenv("GUARDRAILS_API_URL", DEFAULT_API_URL).rstrip("/")
        result = await client.post(
            f"{url}/api/v1/policy/load",
            json={"policy": guardrails},
            headers={
                "Authorization": invariant_authorization,
                "Accept": "application/json",
            },
        )
        result.raise_for_status()


async def preload_guardrails(context: "RequestContext") -> None:
    """
    Preloads the guardrails for faster checking later.

    Args:
        context: RequestContext object.
    """
    if not context.guardrails:
        return

    try:
        # Move these calls to a batch preload/validate API.
        for blocking_guardrail in context.guardrails.blocking_guardrails:
            task = asyncio.create_task(
                _preload(
                    blocking_guardrail.content, context.get_guardrailing_authorization()
                )
            )
            asyncio.shield(task)
        for logging_guadrail in context.guardrails.logging_guardrails:
            task = asyncio.create_task(
                _preload(
                    logging_guadrail.content,
                    context.get_guardrailing_authorization(),
                )
            )
            asyncio.shield(task)
    except Exception as e:
        print(f"Error scheduling preload_guardrails task: {e}")


class ExtraItem:
    """
    Return this class in a instrumented stream callback, to yield an extra item in the resulting stream.
    """

    def __init__(self, value, end_of_stream=False):
        self.value = value
        self.end_of_stream = end_of_stream

    def __str__(self):
        return f"<ExtraItem value={self.value} end_of_stream={self.end_of_stream}>"


class Replacement(ExtraItem):
    """
    Like ExtraItem, but used to replace the full request result in case of 'InstrumentedResponse'.
    """

    def __init__(self, value):
        super().__init__(value, end_of_stream=True)

    def __str__(self):
        return f"<Replacement value={self.value}>"


class InstrumentedStreamingResponse:
    def __init__(self):
        # request statistics
        self.stat_token_times = []
        self.stat_before_time = None
        self.stat_after_time = None

        self.stat_first_item_time = None

    async def on_chunk(self, chunk: Any) -> ExtraItem | None:
        """
        This called will be called on every chunk (async).
        """
        pass

    async def on_start(self) -> ExtraItem | None:
        """
        Decorator to register a listener for start events.
        """
        pass

    async def on_end(self) -> ExtraItem | None:
        """
        Decorator to register a listener for end events.
        """
        pass

    async def event_generator(self):
        """
        Streams the async iterable and invokes all instrumented hooks.

        Args:
            async_iterable: An async iterable to stream.

        Yields:
            The streamed data.
        """
        raise NotImplementedError("This method should be implemented in a subclass.")

    async def instrumented_event_generator(self):
        """
        Streams the async iterable and invokes all instrumented hooks.

        Args:
            async_iterable: An async iterable to stream.

        Yields:
            The streamed data.
        """
        try:
            start = time.time()

            # schedule on_start which can be run concurrently
            start_task = asyncio.create_task(self.on_start(), name="instrumentor:start")

            # create async iterator from async_iterable
            aiterable = aiter(self.event_generator())

            # [STAT] capture start time of first item
            start_first_item_request = time.time()

            # waits for first item of the iterable
            async def wait_for_first_item():
                nonlocal start_first_item_request, aiterable

                r = await aiterable.__anext__()
                if self.stat_first_item_time is None:
                    # [STAT] capture time to first item
                    self.stat_first_item_time = time.time() - start_first_item_request
                return r

            next_item_task = asyncio.create_task(
                wait_for_first_item(), name="instrumentor:next:first"
            )

            # check if 'start_task' yields an extra item
            if extra_item := await start_task:
                # yield extra value before any real items
                yield extra_item.value
                # stop the stream if end_of_stream is True
                if extra_item.end_of_stream:
                    # if first item is already available
                    if not next_item_task.done():
                        # cancel the task
                        next_item_task.cancel()
                        # [STAT] capture time to first item to be now +0.01
                        if self.stat_first_item_time is None:
                            self.stat_first_item_time = (
                                time.time() - start_first_item_request
                            ) + 0.01
                    # don't wait for the first item if end_of stream is True
                    return

            # [STAT] capture before time stamp
            self.stat_before_time = time.time() - start

            while True:
                # wait for first item
                try:
                    item = await next_item_task
                except StopAsyncIteration:
                    break

                # schedule next item
                next_item_task = asyncio.create_task(
                    aiterable.__anext__(), name="instrumentor:next"
                )

                # [STAT] capture token time stamp
                if len(self.stat_token_times) == 0:
                    self.stat_token_times.append(time.time() - start)
                else:
                    self.stat_token_times.append(
                        time.time() - start - sum(self.stat_token_times)
                    )

                if extra_item := await self.on_chunk(item):
                    yield extra_item.value
                    # if end_of_stream is True, stop the stream
                    if extra_item.end_of_stream:
                        # cancel next task
                        next_item_task.cancel()
                        return

                # yield item
                yield item

            # run on_end, before closing the stream (may yield an extra value)
            if extra_item := await self.on_end():
                # yield extra value before any real items
                yield extra_item.value
                # we ignore end_of_stream here, because we are already at the end

            # [STAT] capture after time stamp
            self.stat_after_time = time.time() - start
        finally:
            # [STAT] end all open intervals if not already closed
            if self.stat_after_time is None:
                self.stat_before_time = time.time() - start
            if self.stat_after_time is None:
                self.stat_after_time = 0
            if self.stat_first_item_time is None:
                self.stat_first_item_time = 0

            # print statistics
            token_times_5_decimale = str([f"{x:.5f}" for x in self.stat_token_times])
            print(
                f"[STATS]\n [token times: {token_times_5_decimale} ({len(self.stat_token_times)})]"
            )
            print(f" [before:             {self.stat_before_time:.2f}s] ")
            print(f" [time-to-first-item: {self.stat_first_item_time:.2f}s]")
            print(
                f" [zero-latency:       {' TRUE' if self.stat_before_time < self.stat_first_item_time else 'FALSE'}]"
            )
            print(
                f" [extra-latency:      {self.stat_before_time - self.stat_first_item_time:.2f}s]"
            )
            print(f" [after:              {self.stat_after_time:.2f}s]")
            if len(self.stat_token_times) > 0:
                print(
                    f" [average token time: {sum(self.stat_token_times) / len(self.stat_token_times):.2f}s]"
                )
            print(f" [total: {time.time() - start:.2f}s]")


class InstrumentedResponse(InstrumentedStreamingResponse):
    """
    A class to instrument an async request with hooks for concurrent
    pre-processing and post-processing (input and output guardrailing).
    """

    async def event_generator(self):
        """
        We implement the 'event_generator' as a single item stream,
        where the item is the full result of the request.
        """
        yield await self.request()

    async def request(self):
        """
        This method should be implemented in a subclass to perform the actual request.
        """
        raise NotImplementedError("This method should be implemented in a subclass.")

    async def instrumented_request(self):
        """
        Returns the 'Response' object of the request, after applying all instrumented hooks.
        """
        results = [r async for r in self.instrumented_event_generator()]
        assert len(results) >= 1, "InstrumentedResponse must yield at least one item"

        # we return the last item, in case the end callback yields an extra item. Then,
        # don't return the actual result but the 'end' result, e.g. for output guardrailing.
        return results[-1]


async def check_guardrails(
    messages: List[Dict[str, Any]],
    guardrails: List[Guardrail],
    context: RequestContext,
) -> Dict[str, Any]:
    """
    Checks guardrails on the list of messages.
    This calls the batch check API of the Guardrails service.

    Args:
        messages (List[Dict[str, Any]]): List of messages to verify the guardrails against.
        guardrails (List[Guardrail]): The guardrails to check against.
        invariant_authorization (str): Value of the
                                       invariant-authorization header.

    Returns:
        Dict: Response containing guardrail check results.
    """
    async with httpx.AsyncClient() as client:
        url = os.getenv("GUARDRAILS_API_URL", DEFAULT_API_URL).rstrip("/")
        try:
            result = await client.post(
                f"{url}/api/v1/policy/check/batch",
                json={
                    "messages": messages,
                    "policies": [g.content for g in guardrails],
                    "parameters": context.guardrails_parameters or {}
                },
                headers={
                    "Authorization": context.get_guardrailing_authorization(),
                    "Accept": "application/json"
                },
                timeout=5,
            )
            if not result.is_success:
                if result.status_code == 401:
                    raise HTTPException(
                        status_code=401,
                        detail="The provided Invariant API key is not valid for guardrail checking. Please ensure you are using the correct API key or pass an alternative API key for guardrail checking specifically via the '{}' header.".format(
                            INVARIANT_GUARDRAIL_SERVICE_AUTHORIZATION_HEADER
                        ),
                    )
                raise Exception(
                    f"Guardrails check failed: {result.status_code} - {result.text}"
                )
            guardrails_result = result.json()

            aggregated_errors = {"errors": []}
            for res, guardrail in zip(guardrails_result.get("result", []), guardrails):
                for error in res.get("errors", []):
                    # add each error to the aggregated errors but keep track
                    # of which guardrail it belongs to
                    aggregated_errors["errors"].append(
                        {
                            **error,
                            "guardrail": {
                                "id": guardrail.id,
                                "name": guardrail.name,
                                "content": guardrail.content,
                                "action": guardrail.action,
                            },
                        }
                    )

                # check for any error_message
                if error_message := res.get("error_message"):
                    return {
                        "errors": [
                            {"args": [error_message], "kwargs": {}, "ranges": []}
                        ]
                    }
            return aggregated_errors
        except HTTPException as e:
            raise e
        except Exception as e:
            print(f"Failed to verify guardrails: {e}")
            # make sure runtime errors are also visible in e.g. Explorer
            return {
                "errors": [
                    {
                        "args": ["Gateway: " + str(e)],
                        "kwargs": {},
                        "ranges": ["messages[0].content:L0"],
                    }
                ]
            }
